from typing import Union, Iterable, Optional
from copy import deepcopy

import numpy as np
from numpy import ndarray
from awkward import Record as akRecord

from sigmaepsilon.core import classproperty
from sigmaepsilon.math.linalg import ReferenceFrame as FrameLike
from sigmaepsilon.math.logical import isboolarray, isintegerarray
from sigmaepsilon.math.linalg.sparse import csr_matrix

from ..typing.abcakwrapper import ABC_AkWrapper
from .akwrapper import AkWrapper
from ..space import CartesianFrame, PointCloud
from ..typing import PolyDataProtocol
from ..utils import collect_nodal_data


__all__ = ["PointData"]


def gen_frame(coords: ndarray) -> CartesianFrame:
    return CartesianFrame(dim=coords.shape[1])


class PointData(AkWrapper, ABC_AkWrapper):
    """
    A class to handle data related to the pointcloud of a polygonal mesh.

    The class is technicall a wrapper around an `awkward.Record` instance.

    .. warning::
       Internal variables used during calculations begin with two leading underscores. Try
       to avoid leading double underscores when assigning custom data to a PointData instance,
       unless you are sure, that it is of no importance for the correct behaviour of the
       class instances.

    Parameters
    ----------
    points: numpy.ndarray, Optional
        Coordinates of some points as a 2d NumPy float array. Default is `None`.
    coords: numpy.ndarray, Optional
        Same as `points`. Default is `None`.
    frame: CartesianFrame, Optional
        The coordinate frame the points are understood in. Default is `None`, which
        means the standard global frame (the ambient frame).

    Example
    -------
    >>> from sigmaepsilon.mesh import CartesianFrame, PointData, triangulate
    >>> frame = CartesianFrame(dim=3)
    >>> coords = triangulate(size=(800, 600), shape=(10, 10))[0]
    >>> pd = PointData(coords=coords, frame=frame)
    >>> pd.activity = np.ones((len(pd)), dtype=bool)
    >>> pd.id = np.arange(len(pd))
    """

    _point_cls_ = PointCloud
    _frame_class_ = CartesianFrame
    _attr_map_ = {
        "x": "__x",  # coordinates
        "activity": "__activity",  # activity of the points
        "id": "__id",  # global indices of the points
    }

    def __init__(
        self,
        *args,
        points: Optional[Union[ndarray, None]] = None,
        coords: Optional[Union[ndarray, None]] = None,
        wrap: Optional[Union[akRecord, None]] = None,
        fields: Optional[Union[Iterable, None]] = None,
        frame: Optional[Union[CartesianFrame, None]] = None,
        db: Optional[Union[akRecord, None]] = None,
        container: Optional[Union[PolyDataProtocol, None]] = None,
        **kwargs,
    ):
        if db is not None:
            wrap = db
        elif wrap is None:
            fields = {} if fields is None else fields
            assert isinstance(fields, dict)

            X = None

            if len(args) > 0:
                if isinstance(args[0], np.ndarray):
                    X = args[0]
            else:
                X = points if coords is None else coords

            if X is None:
                raise ValueError("Coordinates must be specified.")

            if not isinstance(X, ndarray):
                raise TypeError("Coordinates must be specified as a NumPy array!")

            fields[self._dbkey_x_] = X
            nP = len(X)

            for k, v in kwargs.items():
                if isinstance(v, np.ndarray):
                    if v.shape[0] == nP:
                        fields[k] = v

        # coordinate frame
        if not isinstance(frame, FrameLike):
            if coords is not None:
                frame = gen_frame(coords)

        self._frame = frame

        super().__init__(*args, wrap=wrap, fields=fields, **kwargs)
        self._container = container

    def __deepcopy__(self, memo):
        return self.__copy__(memo)

    def __copy__(self, memo=None):
        cls = type(self)
        is_deep = memo is not None

        if is_deep:
            copy_function = lambda x: deepcopy(x, memo)
        else:
            copy_function = lambda x: x

        db = copy_function(self.db)
        f = self.frame

        if f is not None:
            axes = copy_function(f.axes)
            if is_deep:
                memo[id(f.axes)] = axes
            frame_cls = type(f)
            frame = frame_cls(axes)
        else:
            frame = None

        result = cls(db=db, frame=frame)

        if is_deep:
            memo[id(self)] = result

        result_dict = result.__dict__

        for k, v in self.__dict__.items():
            if not k in result_dict:
                setattr(result, k, copy_function(v))

        return result

    @classproperty
    def _dbkey_id_(cls) -> str:
        return cls._attr_map_["id"]

    @classproperty
    def _dbkey_x_(cls) -> str:
        return cls._attr_map_["x"]

    @classproperty
    def _dbkey_activity_(cls) -> str:
        return cls._attr_map_["activity"]

    @property
    def has_id(self) -> bool:
        """
        Returns `True` if the points are equipped with IDs, `False` if
        they are not.
        """
        return self._dbkey_id_ in self._wrapped.fields

    @property
    def has_x(self) -> bool:
        """
        Returns `True` if the instance is equipped with coordinates, `False`
        if it isn't.
        """
        return self._dbkey_x_ in self._wrapped.fields

    @property
    def has_activity(self) -> bool:
        """
        Returns `True` if the instance is equipped with activity information, ű
        `False` if it isn't.
        """
        return self._dbkey_activity_ in self._wrapped.fields

    @property
    def container(self) -> PolyDataProtocol:
        """
        Returns the container object of the block.
        """
        return self._container

    @container.setter
    def container(self, value: PolyDataProtocol) -> None:
        """
        Sets the container of the block.
        """
        assert isinstance(value, PolyDataProtocol)
        self._container = value

    @property
    def frame(self) -> FrameLike:
        """
        Returns the frame of the underlying pointcloud.
        """
        result = None

        if isinstance(self._frame, FrameLike):
            result = self._frame

        if result is None:
            dim = self.x.shape[-1]
            result = self._frame_class_(dim=dim)

        return result

    @property
    def activity(self) -> ndarray:
        """
        Returns the activity of the points as an 1d NumPy bool array.
        """
        return self._wrapped[self._dbkey_activity_].to_numpy()

    @activity.setter
    def activity(self, value: ndarray) -> None:
        """
        Sets the activity of the points with an 1d NumPy bool array.
        """
        if not isinstance(value, ndarray):
            raise TypeError(f"Expected a NumPy array, got {type(value)}.")

        if not isboolarray(value):
            raise ValueError(f"Expected a boolean array, got dtype {value.dtype}.")

        if not len(value.shape) == 1:
            raise ValueError("The provided array must be 1 dimensional.")

        if self.has_x and not len(value) == len(self):
            raise ValueError(
                f"The provided array contains {len(value)} values, but there are "
                f"{len(self)} points in the dataset."
            )

        self._wrapped[self._dbkey_activity_] = value

    @property
    def x(self) -> ndarray:
        """
        Returns the coordinates as a 2d NumPy array.
        """
        return self._wrapped[self._dbkey_x_].to_numpy()

    @x.setter
    def x(self, value: ndarray) -> None:
        """
        Sets the coordinates with a 2d NumPy float array.
        """
        if not isinstance(value, ndarray):
            raise TypeError(f"Expected a NumPy array, got {type(value)}")

        if not len(value.shape) == 2:
            raise ValueError("The provided array must be 2 dimensional.")

        self._wrapped[self._dbkey_x_] = value.astype(float)

    @property
    def id(self) -> ndarray:
        """
        Returns the IDs of the points as an 1d NumPy integer array.
        """
        return self._wrapped[self._dbkey_id_].to_numpy()

    @id.setter
    def id(self, value: ndarray) -> None:
        """
        Sets the IDs of the points with an 1d NumPy integer array.
        """
        if not isinstance(value, ndarray):
            raise TypeError(f"Expected a NumPy array, got {type(value)}")

        if not isintegerarray(value):
            raise ValueError(f"Expected an integer array, got dtype {value.dtype}.")

        if not len(value.shape) == 1:
            raise ValueError("The provided array must be 1 dimensional.")

        if self.has_x and not len(value) == len(self):
            raise ValueError(
                f"The provided array contains {len(value)} values, but there are "
                f"{len(self)} points in the dataset."
            )

        self._wrapped[self._dbkey_id_] = value.astype(int)

    def pull(self, key: str, ndf: Union[ndarray, csr_matrix] = None) -> ndarray:
        """
        Pulls data from the cells in the model. The pulled data is either copied or
        distributed according to a measure.

        Parameters
        ----------
        key: str
            A field key to identify data in the databases of the attached
            CellData instances of the blocks.
        ndf: Union[numpy.ndarray, csr_matrix], Optional
            The nodal distribution factors to use. If not provided, the
            default factors are used. Default is None.

        See Also
        --------
        :func:`~sigmaepsilon.mesh.data.polydata.PolyData.nodal_distribution_factors`
        :func:`~sigmaepsilon.mesh.utils.utils.collect_nodal_data`
        """
        source: PolyDataProtocol = self.container.source()

        if ndf is None:
            ndf = source.nodal_distribution_factors()

        if isinstance(ndf, ndarray):
            ndf = csr_matrix(ndf)

        blocks = list(source.cellblocks(inclusive=True))
        b = blocks.pop(0)
        cids = b.cd.id
        topo = b.cd.nodes
        celldata = b.cd.db[key].to_numpy()

        if len(celldata.shape) == 1:
            nE, nNE = topo.shape
            celldata = np.repeat(celldata, nNE).reshape(nE, nNE)

        shp = [len(self)] + list(celldata.shape[2:])
        res = np.zeros(shp, dtype=float)
        collect_nodal_data(celldata, topo, cids, ndf, res)

        for b in blocks:
            cids = b.cd.id
            topo = b.cd.nodes
            celldata = b.cd.db[key].to_numpy()
            if len(celldata.shape) == 1:
                nE, nNE = topo.shape
                celldata = np.repeat(celldata, nNE).reshape(nE, nNE)
            collect_nodal_data(celldata, topo, cids, ndf, res)

        return res



from typing import Tuple

import numpy as np
from numpy import ndarray

from sigmaepsilon.math import atleast2d

from .knn import k_nearest_neighbours
from .utils import cell_centers_bulk
from .cells.utils import (
    _find_first_hits_,
    _find_first_hits_knn_,
    _ntri_to_loc_bulk_,
)
from .tri import _glob_to_nat_tri_bulk_knn_, __pip_tri_bulk__, _glob_to_nat_tri_bulk_
from ..utils.utils import points_of_cells


def locate_tri_2d(
    x: ndarray,
    coords: ndarray,
    triangles: ndarray,
    trimap: ndarray = None,
    lazy: bool = True,
    k: int = 4,
    tol: float = 1e-12,
) -> Tuple[ndarray]:
    x = atleast2d(x, front=True)

    if trimap is None:
        trimap = np.array([[0, 1, 2]], dtype=int)
    lcoords = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]])
    ecoords_tri = points_of_cells(coords, triangles, centralize=False)

    # perform point-in-polygon test for triangles
    if lazy:
        centers_tri = cell_centers_bulk(coords, triangles)
        k_tri = min(k, len(centers_tri))
        neighbours_tri = k_nearest_neighbours(centers_tri, x, k=k_tri)
        nat_tri = _glob_to_nat_tri_bulk_knn_(
            x, ecoords_tri, neighbours_tri
        )  # (nP, kTET, 4)
        pips_tri = __pip_tri_bulk__(nat_tri, tol)  # (nP, kTET)
    else:
        nat_tri = _glob_to_nat_tri_bulk_(x, ecoords_tri)  # (nP, nTET, 4)
        pips_tri = __pip_tri_bulk__(nat_tri, tol)  # (nP, nTET)

    # locate the points that are inside any of the cells
    pip = np.squeeze(np.any(pips_tri, axis=1))  # (nP)
    i_source = np.where(pip)[0]  # (nP_)
    if lazy:
        points_to_tris, points_to_neighbours = _find_first_hits_knn_(
            pips_tri[i_source], neighbours_tri[i_source]
        )
    else:
        points_to_tris, points_to_neighbours = _find_first_hits_(pips_tri[i_source])
    tets_to_cells = np.floor(np.arange(len(triangles)) / len(trimap)).astype(int)
    i_target = tets_to_cells[points_to_tris]  # (nP_)

    # locate the cells that contain the points
    cell_tri_indices = np.tile(np.arange(trimap.shape[0]), len(triangles))[
        points_to_tris
    ]
    nat_tri = nat_tri[i_source]  # (nP_, nTET, 4)
    locations_target = _ntri_to_loc_bulk_(
        lcoords, nat_tri, trimap, cell_tri_indices, points_to_neighbours
    )

    return i_source, i_target, locations_target



from ..config import __hasvtk__
from ..helpers import exporters

if not __hasvtk__:  # pragma: no cover

    def to_vtk(*_) -> None:
        raise ImportError(
            "You need VTK for this. Install it with 'pip install vtk'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    import vtk
    from typing import Union

    from ..data import PolyData
    from ..vtkutils import mesh_to_UnstructuredGrid as mesh_to_vtk

    def to_vtk(
        obj: PolyData, deepcopy: bool = False, multiblock: bool = False
    ) -> Union[vtk.vtkUnstructuredGrid, vtk.vtkMultiBlockDataSet]:
        """
        Returns the mesh as a `VTK` object.

        Parameters
        ----------
        deepcopy: bool, Optional
            Default is False.
        multiblock: bool, Optional
            Wether to return the blocks as a `vtkMultiBlockDataSet` or a list
            of `vtkUnstructuredGrid` instances. Default is False.

        Returns
        -------
        vtk.vtkUnstructuredGrid or vtk.vtkMultiBlockDataSet
        """
        ugrids = []
        for block, c, t, _ in obj._detach_block_data_():
            vtk_cell_id = block.celltype.Geometry.vtk_cell_id
            ugrid = mesh_to_vtk(c, t, vtk_cell_id, deepcopy)
            ugrids.append(ugrid)

        if multiblock:
            mb = vtk.vtkMultiBlockDataSet()
            mb.SetNumberOfBlocks(len(ugrids))

            for i, ugrid in enumerate(ugrids):
                mb.SetBlock(i, ugrid)

            return mb
        else:
            if len(ugrids) > 1:
                return ugrids
            else:
                return ugrids[0]


exporters["vtk"] = to_vtk

__all__ = ["to_vtk"]



import numpy as np
from numbers import Number

from sectionproperties.pre.pre import DEFAULT_MATERIAL, Material
from sectionproperties.pre.geometry import Geometry
from sectionproperties.pre.library.steel_sections import circular_hollow_section as CHS
from sectionproperties.pre.library.steel_sections import (
    rectangular_hollow_section as RHS,
)
from sectionproperties.pre.library.primitive_sections import rectangular_section as RS
from sectionproperties.pre.library.steel_sections import i_section
from sectionproperties.pre.library.steel_sections import tapered_flange_i_section as TFI
from sectionproperties.pre.library.steel_sections import channel_section as PFC
from sectionproperties.pre.library.steel_sections import tapered_flange_channel as TFC
from sectionproperties.pre.library.steel_sections import tee_section
from sectionproperties.analysis.section import Section

from sigmaepsilon.core.wrapping import Wrapper
from sigmaepsilon.core.kwargtools import getallfromkwargs
from sigmaepsilon.mesh.utils import centralize
from sigmaepsilon.mesh.data import TriMesh, PolyData
from sigmaepsilon.mesh.utils.topology import T6_to_T3, detach_mesh_bulk

from ..cells import T3
from ..data import PointData
from ..space import CartesianFrame
from ..utils import coords_to_3d
from ..topoarray import TopologyArray

__all__ = ["generate_mesh", "get_section", "LineSection"]


def generate_mesh(
    geometry: Geometry, *, l_max: float = None, a_max: float = None, n_max: int = None
) -> Geometry:
    """
    Calculates a float describing the maximum mesh element area to be used
    for the finite-element mesh of the section.

    Parameters
    ----------
    geometry: :class:`sectionproperties.pre.geometry.Geometry`
        Describes the shape of the section, i.e. 'I', 'H', 'RHS', etc.
    l_max: float, Optional
        Maximum element edge length. Default is None.
    a_max: float, Optional
        Maximum element area. Default is None.
    n_max: int, Optional
        Maximum number of elements. Default is None.

    Note
    ----
    The value of the mesh size is derived from the assumption of a perfectly
    regular mesh of equilateral triangles.

    Returns
    -------
    :class:`sectionproperties.pre.geometry.Geometry`
        The geometry object provided with a finite element mesh.
    """
    area = geometry.calculate_area()
    mesh_sizes_max = []

    if isinstance(l_max, Number):
        mesh_sizes_max.append(l_max**2 * np.sqrt(3) / 4)

    if isinstance(a_max, Number):
        mesh_sizes_max.append(a_max)

    if isinstance(n_max, int):
        mesh_sizes_max.append(area / n_max)

    mesh_size_max = None
    if len(mesh_sizes_max) > 0:
        mesh_size_max = min(mesh_sizes_max)
    else:
        raise ValueError(
            "Invalid input, the maximum mesh size can't be tedermined!"
        )  # pragma: no cover

    geometry.create_mesh(mesh_sizes=[mesh_size_max])
    return geometry


def get_section(
    shape: str,
    *,
    mesh_params: dict | None = None,
    material: Material | None = None,
    **section_params,
) -> Section:
    """
    Returns a :class:`sectionproperties.analysis.section.Section` instance.

    The parameters in `section_params` are forwarded to the appropriate constructor of the
    `sectionproperties` library. For the available parameters, see their documentation:

        https://sectionproperties.readthedocs.io/en/latest/rst/section_library.html

    Parameters
    ----------
    shape: str
        Describes the shape of the section. The currently supported section types
        are 'I', 'CHS', 'RHS', 'TFI', 'PFC', 'TFC', 'T'.
    mesh_params: dict, Optional
        A dictionary of parameters controlling mesh generation. Default is None.
        For the possible keys and values see :func:`generate_mesh`, to which
        these parameters are forwarded to. Default is None.
    material: :class:`~sectionproperties.pre.pre.Material`, Optional
        The material of the section. If not specified, a default material is
        used. Default is None.
    **section_params: dict
        Parameters required for a given section. See the documentation of
        the `sectionproperties` library for more details. The parameters
        are only required, if the first argument is a string.

    Note
    ----
    Specification of a material is only necessary, if stresses are to be
    calculated. If the reason of creating the section is the calculation of
    geometrical properties of a section, the choice of material is irrelevant.

    Returns
    -------
    :class:`sectionproperties.analysis.section.Section`
        An object representing a cross-section of a beam.

    Examples
    --------
    >>> from sigmaepsilon.mesh.domains.section import get_section
    >>> mesh_params = dict(n_max=500)
    >>> section = get_section('CHS', d=1.0, t=0.1, n=64, mesh_params=mesh_params)
    """

    material = DEFAULT_MATERIAL if material is None else material
    geom, ms = None, None
    if shape == "CHS":
        geom = CHS(
            d=section_params["d"],
            t=section_params["t"],
            n=section_params.get("n", 64),
            material=material,
        )
        ms = section_params["t"]
    elif shape == "RS":
        keys = ["d", "b"]
        params = getallfromkwargs(keys, **section_params)
        params_dict = {k: v for k, v in zip(keys, params)}
        geom = RS(material=material, **params_dict)
        ms = min(params)
    elif shape == "RHS":
        keys = ["d", "b", "t", "r_out", "n_r"]
        params = getallfromkwargs(keys, **section_params)
        params_dict = {k: v for k, v in zip(keys, params)}
        geom = RHS(material=material, **params_dict)
        ms = section_params["t"]
    elif shape == "I":
        keys = ["d", "b", "t_f", "t_w", "r", "n_r"]
        params = getallfromkwargs(keys, **section_params)
        params_dict = {k: v for k, v in zip(keys, params)}
        geom = i_section(material=material, **params_dict)
        ms = min(section_params["t_f"], section_params["t_w"])
    elif shape == "TFI":
        keys = ["d", "b", "t_f", "t_w", "r_r", "r_f", "alpha", "n_r"]
        params = getallfromkwargs(keys, **section_params)
        params_dict = {k: v for k, v in zip(keys, params)}
        geom = TFI(material=material, **params_dict)
        ms = min(section_params["t_f"], section_params["t_w"])
    elif shape == "PFC":
        keys = ["d", "b", "t_f", "t_w", "r", "n_r"]
        params = getallfromkwargs(keys, **section_params)
        params_dict = {k: v for k, v in zip(keys, params)}
        geom = PFC(material=material, **params_dict)
        ms = min(section_params["t_f"], section_params["t_w"])
    elif shape == "TFC":
        keys = ["d", "b", "t_f", "t_w", "r_r", "r_f", "alpha", "n_r"]
        params = getallfromkwargs(keys, **section_params)
        params_dict = {k: v for k, v in zip(keys, params)}
        geom = TFC(material=material, **params_dict)
        ms = min(section_params["t_f"], section_params["t_w"])
    elif shape == "T":
        keys = ["d", "b", "t_f", "t_w", "r", "n_r"]
        params = getallfromkwargs(keys, **section_params)
        params_dict = {k: v for k, v in zip(keys, params)}
        geom = tee_section(material=material, **params_dict)
        ms = min(section_params["t_f"], section_params["t_w"])
    else:
        raise NotImplementedError(
            "Section type <{}> is not yet implemented :(".format(shape)
        )

    if geom is not None:
        if mesh_params is None:
            assert ms is not None, "Invalid input!"
            mesh_params = dict(l_max=ms)
        assert isinstance(mesh_params, dict)
        geom = generate_mesh(geom, **mesh_params)
        return Section(geom)
    raise RuntimeError("Unable to get section.")  # pragma: no cover


class LineSection(Wrapper):
    """
    Wraps an instance of `sectionproperties.analysis.section.Section` and
    adds a little here and there to make some of the functionality more
    accessible.

    Parameters
    ----------
    *args: tuple, Optional
        The first parameter can be a string referring to a section type, or an
        instance of :class:`sectionproperties.analysis.section.Section`. In the
        former case, parameters of the section can be provided as keyword arguments,
        which are then forwarded to :func:`get_section`, see its documentation
        for further details.
    **kwargs: dict, Optional
        The parameters of a section as keyword arguments, only if the first
        positional argument is a string (see above).
    mesh_params: dict, Optional
        A dictionary controlling the density of the mesh of the section.
        Default is None.
    material: :class:`sectionproperties.pre.pre.Material`, Optional
        The material of the section. If not specified, a default material is
        used. Default is None.
    wrap: :class:`sectionproperties.analysis.section.Section`
        A section object to be wrapped. It can also be provided as the
        first positional argument.
    Notes
    -----
    The implementation here only covers homogeneous cross sections. If you want
    to define an inhomogeneous section, it must be explicity given either as
    the first position argument, or with the keyword argument `wrap`.

    Examples
    --------
    >>> from sigmaepsilon.mesh.domains.section import LineSection, get_section
    >>> section = LineSection(get_section('CHS', d=1.0, t=0.1, n=64))

    or simply provide the shape as the first argument and everything
    else with keyword arguments:

    >>> section = LineSection('CHS', d=1.0, t=0.1, n=64)

    Example
    -------
    .. plot::
        :include-source: True

        Plot a section with Matplotlib using 3-noded triangles:

        import matplotlib.pyplot as plt
        from sigmaepsilon.mesh.plotting.mpl import triplot_mpl_mesh
        section = LineSection('CHS', d=1.0, t=0.3, n=32,
                              mesh_params=dict(n_max=20))
        triobj = section.trimesh(order=1).to_triobj()
        fig, ax = plt.subplots(figsize=(4, 2))
        triplot_mpl_mesh(triobj, fig=fig, ax=ax, lw=0.1)
    """

    def __init__(
        self,
        *args,
        wrap=None,
        shape=None,
        mesh_params=None,
        material: Material = None,
        **kwargs,
    ):
        if len(args) > 0:
            try:
                if isinstance(args[0], str):
                    wrap = get_section(
                        args[0], mesh_params=mesh_params, material=material, **kwargs
                    )
                else:
                    if shape is None:
                        if isinstance(args[0], Section):
                            wrap = args[0]
                    else:
                        wrap = get_section(
                            shape, mesh_params=mesh_params, material=material, **kwargs
                        )
            except Exception as e:  # pragma: no cover
                raise RuntimeError(f"Unable to create section: {e}")

        super().__init__(*args, wrap=wrap, **kwargs)
        self.props = None

    def coords(self) -> np.ndarray:
        """
        Returns centralized vertex coordinates of the supporting
        point cloud as a numpy array.
        """
        return centralize(np.array(self.mesh["vertices"]))

    def topology(self) -> TopologyArray:
        """
        Returns vertex indices of T6 triangles as a numpy array.
        """
        return TopologyArray(np.array(self.mesh["triangles"].tolist()))

    def trimesh(self, subdivide: bool = False, order: int = 1, **kwargs) -> TriMesh:
        """
        Returns the mesh of the section as a collection of T3 triangles.
        Keyword arguments are forwarded to the constructor of
        :class:`~sigmaepsilon.mesh.tri.trimesh.TriMesh`.

        Parameters
        ----------
        order: boolean, Optional
            Order of the tetrahedra. Order 1 means linear, order 2 quadratic. Default is 1.
        subdivide: boolean, Optional
            Controls how the T6 triangles are transformed into T3 triangles,
            if the argument 'T6' is False. If True, the T6 triangles
            are subdivided into 4 T3 triangles. If False, T3 triangles are
            formed by the corners of T6 triangles only, and all the remaining
            nodes are neglected.

        See Also
        --------
        :class:`~sigmaepsilon.mesh.trimesh.TriMesh`

        Examples
        --------
        >>> from sigmaepsilon.mesh.domains import LineSection
        >>> section = LineSection(get_section('CHS', d=1.0, t=0.1, n=64))
        >>> trimesh = section.trimesh()
        """
        triangles = self.topology()
        assert (
            not triangles.is_jagged()
        ), "Only homogeneous topologies are supported at the moment."
        points, triangles = coords_to_3d(self.coords()), triangles.to_numpy()
        if order == 1:
            if subdivide:
                path = np.array([[0, 5, 4], [5, 1, 3], [3, 2, 4], [5, 3, 4]], dtype=int)
                points, triangles = T6_to_T3(points, triangles, path=path)
            else:
                points, triangles = detach_mesh_bulk(points, triangles[:, :3])
        else:  # pragma: no cover
            raise NotImplementedError("Higher order elements are not yet supported!")

        frame = kwargs.get("frame", CartesianFrame(dim=3))
        pd = PointData(coords=points, frame=frame)
        cd = T3(topo=triangles, frames=frame)
        return TriMesh(pd, cd)

    def extrude(
        self,
        *,
        length: float,
        frame: CartesianFrame,
        N: int,
        **__,
    ) -> PolyData:
        """
        Creates a 3d tetragonal mesh from the section.

        Parameters
        ----------
        length: float
            Length of the beam.
        N: int
            Number of subdivisions along the length of the beam.
        frame: CartesianFrame
            An object instance representing an orthonormal coordinate frame.

        Returns
        -------
        :class:`~sigmaepsilon.mesh.data.polydata.PolyData`
        """
        return self.trimesh(frame=frame).extrude(h=length, N=N)

    def calculate_geometric_properties(self, *args, **kwargs):
        return self._wrapped.calculate_geometric_properties(*args, **kwargs)

    def calculate_warping_properties(self, *args, **kwargs):
        return self._wrapped.calculate_warping_properties(*args, **kwargs)

    @property
    def A(self) -> float:
        """
        Returns the cross-sectional area.
        """
        return self.section_props.area

    @property
    def Ix(self) -> float:
        """
        Returns the second moment of inertia around 'x'.
        """
        return self.Iy + self.Iz

    @property
    def Iy(self) -> float:
        """
        Returns the second moment of inertia around 'y'.
        """
        return self.section_props.ixx_c

    @property
    def Iz(self) -> float:
        """
        Returns the second moment of inertia around 'z'.
        """
        return self.section_props.iyy_c

    @property
    def geometric_properties(self) -> dict:
        """
        Returns the geometric properties of the section.
        """
        return {"A": self.A, "Ix": self.Ix, "Iy": self.Iy, "Iz": self.Iz}

    @property
    def section_properties(self) -> dict:
        """
        Returns all properties of the section.
        """
        return self.props

    def calculate_section_properties(self) -> dict:
        """
        Retruns a dictionary containing the properties of the section.
        """
        self.calculate_geometric_properties()
        section_properties = self.geometric_properties
        self.calculate_warping_properties()
        self.props = section_properties
        return section_properties

    def get_section_properties(self) -> dict:
        """
        Returns all properties of the section.
        """
        if self.props is not None:
            return self.props
        else:
            return self.calculate_section_properties()



from ..config import __haspyvista__
from ..helpers import exporters

if not __haspyvista__:  # pragma: no cover

    def to_pv(*_) -> None:
        raise ImportError(
            "You need PyVista for this. Install it with 'pip install pyvista'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    from typing import Union, Optional
    from contextlib import suppress

    import pyvista as pv
    import vtk
    from numpy import ndarray

    from ..data import PolyData
    from ..vtkutils import mesh_to_UnstructuredGrid as mesh_to_vtk

    pyVistaLike = Union[pv.PolyData, pv.PointGrid, pv.UnstructuredGrid]

    def to_pv(
        obj: PolyData,
        deepcopy: Optional[bool] = False,
        multiblock: Optional[bool] = False,
        scalars: Optional[Union[str, ndarray, None]] = None,
    ) -> Union[pv.UnstructuredGrid, pv.MultiBlock]:
        """
        Returns the mesh as a `PyVista` object, optionally set up with data.

        Parameters
        ----------
        deepcopy: bool, Optional
            Default is False.
        multiblock: bool, Optional
            Wether to return the blocks as a `vtkMultiBlockDataSet` or a list
            of `vtkUnstructuredGrid` instances. Default is False.
        scalars: str or numpy.ndarray, Optional
            A string or an array describing scalar data. Default is None.

        Returns
        -------
        pyvista.UnstructuredGrid or pyvista.MultiBlock
        """
        ugrids = []
        data = []
        for block, c, t, d in obj._detach_block_data_(scalars):
            vtk_cell_id = block.celltype.Geometry.vtk_cell_id
            ugrid = mesh_to_vtk(c, t, vtk_cell_id, deepcopy)
            ugrids.append(ugrid)
            data.append(d)

        if multiblock:
            mb = vtk.vtkMultiBlockDataSet()
            mb.SetNumberOfBlocks(len(ugrids))

            for i, ugrid in enumerate(ugrids):
                mb.SetBlock(i, ugrid)

            mb = pv.wrap(mb)

            with suppress(AttributeError):
                mb.wrap_nested()

            return mb
        else:
            if scalars is None:
                return [pv.wrap(ugrid) for ugrid in ugrids]
            else:
                res = []
                for ugrid, d in zip(ugrids, data):
                    pvobj = pv.wrap(ugrid)
                    if isinstance(d, ndarray):
                        if isinstance(scalars, str):
                            pvobj[scalars] = d
                        else:
                            pvobj["scalars"] = d
                    res.append(pvobj)
                return res


exporters["PyVista"] = to_pv

__all__ = ["to_pv"]



# -*- coding: utf-8 -*-
import numpy as np
from numba import njit

from sigmaepsilon.math import tile

from .utils.tri import edges_tri
from .utils.topology import unique_topo_data

__cache = True


def populate_trimesh_T3(
    points: np.ndarray,
    topo: np.ndarray,
    dp: np.ndarray,
    N=1,
    return_lines=False,
    edges=None,
    edgeIDs=None,
):
    points_pop, topo_pop = populate_trimesh_T3_njit(points, topo, dp, N)
    if return_lines:
        if edges is None or edgeIDs is None:
            edges, edgeIDs = unique_topo_data(edges_tri(topo))
        nEdge = len(edges)
        nP = len(points)
        edges_pop = tile(edges, nP, N)
        edgeIDs_pop = tile(edgeIDs, nEdge, N)
        return points_pop, edges_pop, topo_pop, edgeIDs_pop
    else:
        return points_pop, topo_pop


@njit(nogil=True, cache=__cache)
def populate_trimesh_T3_njit(points: np.ndarray, topo: np.ndarray, dp: np.ndarray, N=1):
    points_pop = tile(points, dp, N)
    topo_pop = tile(topo, len(points), N)
    return points_pop, topo_pop



from typing import (
    Callable,
    Optional,
    ClassVar,
    Iterable,
    Tuple,
    List,
    Union,
)
from abc import abstractmethod

import numpy as np
from numpy import ndarray
from sympy import Matrix, lambdify, symbols

from sigmaepsilon.core.meta import ABCMeta_Weak
from sigmaepsilon.math import atleast1d, atleast2d, ascont
from sigmaepsilon.math.utils import to_range_1d

from ..typing import GeometryProtocol
from ..utils import cell_center, cell_center_2d
from ..utils.utils import global_shape_function_derivatives
from ..cellapproximator import LagrangianCellApproximator
from ..triang import triangulate


__all__ = [
    "PolyCellGeometry1d",
    "PolyCellGeometry2d",
    "PolyCellGeometry3d",
]


class ABC(metaclass=ABCMeta_Weak):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance.
    """

    __slots__ = ()


class PolyCellGeometry(ABC):
    """
    Base class for classes that implement the geometry protocol.

    See Also
    --------
    :class:`~sigmaepsilon.mesh.typing.geometry.GeometryProtocol`
    """

    number_of_nodes: ClassVar[int]
    number_of_spatial_dimensions: ClassVar[int]
    number_of_nodal_variables = 1
    vtk_cell_id: ClassVar[Optional[int]] = None
    meshio_cell_id: ClassVar[Optional[str]] = None
    boundary_class: ClassVar[Optional["PolyCellGeometry"]] = None
    shape_function_evaluator: ClassVar[Optional[Callable]] = None
    shape_function_matrix_evaluator: ClassVar[Optional[Callable]] = None
    shape_function_derivative_evaluator: ClassVar[Optional[Callable]] = None
    monomial_evaluator: ClassVar[Optional[Callable]] = None
    quadrature: ClassVar[Optional[dict]] = None

    @classmethod
    def generate_class(
        cls, base: Optional[Union[GeometryProtocol, None]] = None, **kwargs
    ) -> GeometryProtocol:
        """
        A factory function that returns a custom 1d class.

        Parameters
        ----------
        base: GeometryProtocol, Optional
            A base class that implements the GeometryProtocol. If not
            provided, the class serves as a base.
        **kwargs: dict, Optional
            A dictionary of class attributes and their values.

        Notes
        -----
        During generation, the generated class only inherits some properties,
        while others are set to default values to avoid inconsistent attributes.

        Example
        -------
        Define a custom 1d cell with 4 nodes:

        >>> from sigmaepsilon.mesh.geometry import PolyCellGeometry1d
        >>> CustomClass = PolyCellGeometry1d.generate_class(number_of_nodes=4)

        This is equivalent to:

        >>> class CustomClass(PolyCellGeometry1d):
        ...     number_of_nodes = 4
        """
        base = cls if not base else base

        class CustomClass(base):
            vtk_cell_id = None
            meshio_cell_id = None
            boundary_class = None
            shape_function_evaluator = None
            shape_function_matrix_evaluator = None
            shape_function_derivative_evaluator = None
            monomial_evaluator = None
            quadrature = None

        for key, value in kwargs.items():
            setattr(CustomClass, key, value)

        return CustomClass

    @classmethod
    @abstractmethod
    def master_coordinates(cls) -> ndarray:
        """
        Returns the coordinates of the master element.

        Returns
        -------
        numpy.ndarray
        """
        ...

    def master_center(cls) -> ndarray:
        """
        Returns the coordinates of the master element.

        Returns
        -------
        numpy.ndarray
        """
        return cell_center(cls.master_coordinates())

    @classmethod
    def shape_function_values(
        cls, x: Union[float, Iterable[float]], *, rng: Iterable = None
    ) -> ndarray:
        """
        Evaluates the shape functions at the specified locations.

        Parameters
        ----------
        x: Union[float, Iterable[float]]
            Locations of the evaluation points in the master domain.
        rng: Iterable, Optional
            The range in which the locations ought to be understood, only for 1d
            cells. Typically [0, 1] or [-1, 1]. Default is [0, 1].

        Returns
        -------
        numpy.ndarray
            An array of shape (nP, nNE) where nP and nNE are the number of
            evaluation points and shape functions. If there is only one
            evaluation point, the returned array is one dimensional.
        """
        if cls.number_of_spatial_dimensions == 1:
            rng = np.array([-1, 1]) if rng is None else np.array(rng)
            x = atleast1d(np.array(x))
            x = to_range_1d(x, source=rng, target=[-1, 1])
        else:
            x = np.array(x)

        if cls.shape_function_evaluator is None:
            cls.generate_class_functions(update=True)

        if cls.number_of_spatial_dimensions == 3:
            if len(x.shape) == 1:
                x = atleast2d(x, front=True)
                return cls.shape_function_evaluator(x).astype(float)

        return cls.shape_function_evaluator(x).astype(float)

    @classmethod
    def shape_function_derivatives(
        cls,
        x: Union[float, Iterable[float]],
        *,
        jac: ndarray = None,
        dshp: ndarray = None,
        rng: Iterable = None,
    ) -> ndarray:
        """
        Evaluates shape function derivatives wrt. the master element or the local
        coordinate frames of some cells. To control the behaviour, either 'jac' or 'wrt'
        can be provided.

        Parameters
        ----------
        x: Union[float, Iterable[float]]
            Locations of the evaluation points.
        jac: Iterable, Optional
            The jacobian matrix as a float array of shape (nE, nP, nD, nD), evaluated for
            an nP number of points and nP number cells and nD number of spatial dimensions.
            Default is None.
        rng: Iterable, Optional
            The range in which the locations ought to be understood, only for 1d
            cells. Typically [0, 1] or [-1, 1]. Default is [0, 1].
        dshp: numpy.ndarray, Optional
            Shape function derivatives wrt. the master element. Only relevant if 'jac' is
            provided. The purpose of this argument is to avoid repeated evaluation in situations
            where 'dshp' is required on its own and is already at hand when calling this function.
            Default is None, in which case it is calculated automatically.

        Notes
        -----
        Only first derivatives are calculated.

        Returns
        -------
        numpy.ndarray
            An array of shape (nP, nNE, nD), where nP, nNE and nD are
            the number of evaluation points, nodes and spatial dimensions.
            If 'jac' is provided, the result is of shape (nE, nP, nNE, nD),
            where nE is the number of cells in the block.
        """
        if x is not None:
            if cls.number_of_spatial_dimensions == 1:
                rng = np.array([-1, 1]) if rng is None else np.array(rng)
                x = atleast1d(np.array(x))
                x = to_range_1d(x, source=rng, target=[-1, 1])
            else:
                x = np.array(x)

        if jac is None:
            x = np.array(x) if x is not None else cls.master_coordinates()

            if cls.shape_function_derivative_evaluator is None:
                cls.generate_class_functions(update=True)

            if cls.number_of_spatial_dimensions == 3:
                if len(x.shape) == 1:
                    x = atleast2d(x, front=True)
                    return cls.shape_function_derivative_evaluator(x).astype(float)

            return cls.shape_function_derivative_evaluator(x).astype(float)
        else:
            x = np.array(x) if x is not None else cls.master_coordinates()

            if dshp is None:
                dshp = cls.shape_function_derivatives(x)

            return global_shape_function_derivatives(dshp, jac)

    @classmethod
    def shape_function_matrix(
        cls, x: Union[float, Iterable[float]], *, rng: Iterable = None, N: int = None
    ) -> ndarray:
        """
        Evaluates the shape function matrix at the specified locations.

        Parameters
        ----------
        x: Union[float, Iterable[float]]
            Locations of the evaluation points.
        rng: Iterable, Optional
            The range in which the locations ought to be understood, only for 1d
            cells. Typically [0, 1] or [-1, 1]. Default is [0, 1].
        N: int, Optional
            Number of unknowns per node.

        Returns
        -------
        numpy.ndarray
            An array of shape (nP, nNE, N * nNE) where nP, nDOF and nNE
            are the number of evaluation points, degrees of freedom per node
            and nodes per cell.
        """
        if cls.number_of_spatial_dimensions == 1:
            rng = np.array([-1, 1]) if rng is None else np.array(rng)
            x = atleast1d(np.array(x))
            x = to_range_1d(x, source=rng, target=[-1, 1])
        else:
            x = np.array(x)

        if cls.shape_function_matrix_evaluator is None:
            cls.generate_class_functions(update=True)

        if cls.number_of_spatial_dimensions == 3:
            if len(x.shape) == 1:
                x = atleast2d(x, front=True)
                if N:
                    return cls.shape_function_matrix_evaluator(x, N).astype(float)
                else:
                    return cls.shape_function_matrix_evaluator(x).astype(float)

        if N:
            return cls.shape_function_matrix_evaluator(x, N).astype(float)
        else:
            return cls.shape_function_matrix_evaluator(x).astype(float)

    def polybase(cls) -> Tuple[List]:
        """
        Ought to retrun the polynomial base of the master element.

        Returns
        -------
        list
            A list of SymPy symbols.
        list
            A list of monomials.
        """
        raise NotImplementedError

    @classmethod
    def generate_class_functions(
        cls,
        return_symbolic: Optional[bool] = True,
        update: Optional[bool] = True,
    ) -> Tuple:
        """
        Generates functions to evaulate shape functions, their derivatives
        and the shape function matrices using SymPy. For this to work, the
        'polybase' and 'lcoords' class methods must be implemented.

        Parameters
        ----------
        return_symbolic: bool, Optional
            If True, the function returns symbolic expressions of shape functions
            and their derivatives. Default is True.
        update: bool, Optional
            If True, class methods are updated with the generated versions.
            Default is True.

        Notes
        -----
        Some cells are equipped with functions to evaluate shape functions a-priori,
        other classes rely on symbolic generation of these functions. In the latter case,
        this function is automatically invoked runtime, there is no need to manually
        trigger it.

        Example
        -------
        >>> from sigmaepsilon.mesh.cells import H8
        >>> shp, dshp, shpf, shpmf, dshpf = H8.Geometry.generate_class_functions()

        Here `shp` and `dshp` are simbolic matrices for shape functions and
        their first derivatives, `shpf`, `shpmf` and `dshpf` are functions for
        fast evaluation of shape function values, the shape function matrix and
        shape function derivatives, respectively.
        """
        nN = cls.number_of_nodes
        nD = cls.number_of_spatial_dimensions
        nX = cls.number_of_nodal_variables
        locvars, monoms = cls.polybase()
        monoms.pop(0)
        lcoords = cls.master_coordinates()
        if nD == 1:
            lcoords = np.reshape(lcoords, (nN, 1))

        def subs(lpos):
            return {v: lpos[i] for i, v in enumerate(locvars)}

        def mval(lpos):
            return [m.evalf(subs=subs(lpos)) for m in monoms]

        M = np.ones((nN, nN), dtype=float)
        M[:, 1:] = np.vstack([mval(loc) for loc in lcoords])
        coeffs = np.linalg.inv(M)
        monoms.insert(0, 1)
        shp = Matrix([np.dot(coeffs[:, i], monoms) for i in range(nN)])
        dshp = Matrix([[f.diff(m) for m in locvars] for f in shp])
        _shpf = lambdify([locvars], shp[:, 0].T, "numpy")
        _dshpf = lambdify([locvars], dshp, "numpy")

        def shpf(p: ndarray) -> ndarray:
            """
            Evaluates the shape functions at multiple points in the
            master domain.
            """
            p = atleast2d(p, back=True)
            r = np.stack([_shpf(p[i])[0] for i in range(len(p))])
            return ascont(r)

        def shpmf(p: ndarray, ndof: int = nX) -> ndarray:
            """
            Evaluates the shape function matrix at multiple points
            in the master domain.
            """
            p = atleast2d(p, back=True)
            nP = p.shape[0]
            eye = np.eye(ndof, dtype=float)
            shp = shpf(p)
            res = np.zeros((nP, ndof, nN * ndof), dtype=float)
            for iP in range(nP):
                for i in range(nN):
                    res[iP, :, i * ndof : (i + 1) * ndof] = eye * shp[iP, i]
            return ascont(res)

        def dshpf(p: ndarray) -> ndarray:
            """
            Evaluates the shape function derivatives at multiple points
            in the master domain.
            """
            p = atleast2d(p, back=True)
            r = np.stack([_dshpf(p[i]) for i in range(len(p))])
            return ascont(r)

        if update:
            cls.shape_function_evaluator = shpf
            cls.shape_function_matrix_evaluator = shpmf
            cls.shape_function_derivative_evaluator = dshpf

        if return_symbolic:
            return shp, dshp, shpf, shpmf, dshpf
        else:
            return shpf, shpmf, dshpf

    @classmethod
    def approximator(cls, x: Iterable = None) -> LagrangianCellApproximator:
        """
        Returns a callable object that can be used to approximate over
        nodal values of one or more cells.

        Parameters
        ----------
        x: Iterable, Optional
            The locations of known data. It can be fed into the returned approximator
            function directly, but since the operation involves the inversion of a matrix
            related to these locations, it is a good idea to pre calculate it if you want
            to reuse the approximator with the same source coordinates.

        Returns
        -------
        :class:`~sigmaepsilon.mesh.cellapproximator.LagrangianCellApproximator`
            A callable approximator class. Refer to its documentation for more examples.

        Notes
        -----
        If the number of source coorindates does not match the number of nodes (and hence
        the number of shape functions) of the master element of the class, the interpolation
        is gonna be under or overdetermined and the operation involves the calculation of a
        generalized inverse.

        See also
        --------
        :class:`~sigmaepsilon.mesh.cellapproximator.LagrangianCellApproximator`

        Examples
        --------
        Let assume that we know some data at some locations:
        >>> source_data = [1, 2, 3, 4]
        >>> source_location = [[-0.5, -0.5], [0.5, -0.5], [0.5, 0.5], [-0.5, 0.5]]

        We want to extrapolate from this information to the location
        >>> target_location = [[-2, -2], [2, -2], [2, 2], [-2, 2]]

        We have provided four points and four data values. If we want an exact extrapolation,
        we use 4-noded quadrilaterals:

        >>> from sigmaepsilon.mesh.cells import Q4
        >>> approximator = Q4.Geometry.approximator()
        >>> target_data = approximator(source=source_location, values=source_data, target=target_location)

        Here we provided 3 inputs to the approximator. If we want to reuse the approximator
        with the same source locations, it is best to provide them when creating the approximator.
        This saves some time for repeated evaluations.

        >>> from sigmaepsilon.mesh.cells import Q4
        >>> approximator = Q4.Geometry.approximator(source_location)
        >>> target_data = approximator(values=source_data, target=target_location)
        """
        return LagrangianCellApproximator(cls, x)


class PolyCellGeometry1d(PolyCellGeometry):
    number_of_spatial_dimensions = 1

    @classmethod
    def polybase(cls) -> Tuple[List]:
        """
        Retruns the polynomial base of the master element.

        Returns
        -------
        list
            A list of SymPy symbols.
        list
            A list of monomials.
        """
        if not isinstance(cls.number_of_nodes, int):
            raise ValueError(
                "Attribute 'number_of_nodes' of the cell must be set to a positive integer"
            )
        locvars = r = symbols("r", real=True)
        monoms = [r**i for i in range(cls.number_of_nodes)]
        return [locvars], monoms

    @classmethod
    def master_coordinates(cls) -> ndarray:
        """
        Returns local coordinates of the cell.

        Returns
        -------
        numpy.ndarray
        """
        if not isinstance(cls.number_of_nodes, int):
            raise ValueError(
                "Attribute 'number_of_nodes' of the cell must be set to a positive integer"
            )
        return np.linspace(-1.0, 1.0, cls.number_of_nodes)

    @classmethod
    def master_center(cls) -> ndarray:
        """
        Returns the local coordinates of the center of the cell.

        Returns
        -------
        numpy.ndarray
        """
        return np.array([0.0])


class PolyCellGeometry2d(PolyCellGeometry):
    number_of_spatial_dimensions = 2

    @classmethod
    def master_center(cls) -> ndarray:
        """
        Ought to return the local coordinates of the center of the
        master element.

        Returns
        -------
        numpy.ndarray
        """
        return cell_center_2d(cls.master_coordinates())

    @classmethod
    def trimap(cls) -> Iterable:
        """
        Returns a mapper to transform topology and other data to
        a collection of triangular simplices.
        """
        _, t, _ = triangulate(points=cls.master_coordinates())
        return t


class PolyCellGeometry3d(PolyCellGeometry):
    number_of_spatial_dimensions = 3

    @classmethod
    def tetmap(cls) -> Iterable:
        """
        Returns a mapper to transform topology and other data to
        a collection of tetrahedral simplices.
        """
        raise NotImplementedError



from numba import njit, prange
import numpy as np
from numpy import ndarray

__cache = True


@njit(nogil=True, cache=__cache)
def monoms_TET4_single(x: ndarray) -> ndarray:
    r, s, t = x
    res = np.array([1, r, s, t, r * s, r * t, s * t, r**2, s**2, t**2], dtype=x.dtype)
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def monoms_TET4_multi(x: ndarray) -> ndarray:
    nP = x.shape[0]
    res = np.zeros((nP, 4), dtype=x.dtype)
    for i in prange(nP):
        res[i] = monoms_TET4_single(x[i])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def monoms_TET4_bulk_multi(x: ndarray) -> ndarray:
    nE = x.shape[0]
    res = np.zeros((nE, 4, 4), dtype=x.dtype)
    for i in prange(nE):
        res[i] = monoms_TET4_multi(x[i])
    return res


def monoms_TET4(x: ndarray) -> ndarray:
    N = len(x.shape)
    if N == 1:
        return monoms_TET4_single(x)
    elif N == 2:
        return monoms_TET4_multi(x)
    elif N == 3:
        return monoms_TET4_bulk_multi(x)
    else:
        raise NotImplementedError


@njit(nogil=True, cache=__cache)
def shp_TET4(pcoord: ndarray) -> ndarray:
    r, s, t = pcoord
    return np.array([1 / 3 - r - s - t, r + 1 / 3, s + 1 / 3, t + 1 / 3])


@njit(nogil=True, parallel=True, cache=__cache)
def shp_TET4_multi(pcoords: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, 4), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP, :] = shp_TET4(pcoords[iP])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_matrix_TET4(pcoord: ndarray, ndof: int = 3) -> ndarray:
    eye = np.eye(ndof, dtype=pcoord.dtype)
    shp = shp_TET4(pcoord)
    res = np.zeros((ndof, ndof * 4), dtype=pcoord.dtype)
    for i in prange(4):
        res[:, i * ndof : (i + 1) * ndof] = eye * shp[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_matrix_TET4_multi(pcoords: ndarray, ndof: int = 3) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, ndof, ndof * 4), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = shape_function_matrix_TET4(pcoords[iP], ndof)
    return res


@njit(nogil=True, cache=__cache)
def dshp_TET4(x: float) -> ndarray:
    res = np.array(
        [[-1.0, -1.0, -1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def dshp_TET4_multi(pcoords: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, 4, 3), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = dshp_TET4(pcoords[iP])
    return res



from numba import njit, prange
import numpy as np
from numpy import ndarray

__cache = True


@njit(nogil=True, cache=__cache)
def monoms_W18_single(x: ndarray) -> ndarray:
    r, s, t = x
    res = np.array(
        [
            1,
            r,
            s,
            r**2,
            s**2,
            r * s,
            t,
            t * r,
            t * s,
            t * r**2,
            t * s**2,
            t * r * s,
            t**2,
            t**2 * r,
            t**2 * s,
            t**2 * r**2,
            t**2 * s**2,
            t**2 * r * s,
        ],
        dtype=x.dtype,
    )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def monoms_W18_multi(x: ndarray) -> ndarray:
    nP = x.shape[0]
    res = np.zeros((nP, 18), dtype=x.dtype)
    for i in prange(nP):
        res[i] = monoms_W18_single(x[i])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def monoms_W18_bulk_multi(x: ndarray) -> ndarray:
    nE = x.shape[0]
    res = np.zeros((nE, 18, 18), dtype=x.dtype)
    for i in prange(nE):
        res[i] = monoms_W18_multi(x[i])
    return res


def monoms_W18(x: ndarray) -> ndarray:
    N = len(x.shape)
    if N == 1:
        return monoms_W18_single(x)
    elif N == 2:
        return monoms_W18_multi(x)
    elif N == 3:
        return monoms_W18_bulk_multi(x)
    else:
        raise NotImplementedError



from .polydata import PolyData
from .pointdata import PointData
from .celldata import CellData
from .linedata import LineData
from .polycell import PolyCell
from .trimesh import TriMesh

__all__ = [
    "PointData",
    "CellData",
    "PolyData",
    "LineData",
    "PolyCell",
    "TriMesh",
]



from typing import Iterable, Union, Any, Callable
from functools import partial
import warnings

import numpy as np
from numpy import ndarray

from sigmaepsilon.core.warning import SigmaEpsilonPerformanceWarning
from sigmaepsilon.math.linalg import generalized_inverse

from .utils.cells.approximator import _approximate_multi

__all__ = ["LagrangianCellApproximator"]


def _get_shape_function_evaluator(cls: Any) -> Callable:
    try:
        if hasattr(cls, "Geometry"):
            shp_fnc = cls.Geometry.shape_function_values
        else:
            shp_fnc = cls.shape_function_values
        return shp_fnc
    except AttributeError:
        raise TypeError(
            "Invalid type. The cell must be an instance of PolyCell"
            " or implement the PolyCellGeometry protocol."
        )


def _approximator(
    cls: Any,
    *,
    x_source: Iterable | None = None,
    shp_source_inverse: Iterable | None = None,
    values_source: Iterable = None,
    x_target: Iterable | None = None,
    axis: int | None = None,
) -> Union[float, ndarray]:
    """
    Returns interpolated values from a set of known points and values.
    """
    shp_fnc: Callable = _get_shape_function_evaluator(cls)

    if shp_source_inverse is None:
        assert isinstance(x_source, Iterable)
        shp_source = shp_fnc(x_source)  # (nP_source, nNE)

        num_rows, num_columns = shp_source.shape
        rank = np.linalg.matrix_rank(shp_source)
        square_and_full_rank = (
            num_rows == num_columns
        ) and rank == num_columns == num_rows
        if not square_and_full_rank:  # pragma: no cover
            warnings.warn(
                "The approximation involves the calculation of a generalized inverse "
                "which probably results in loss of precision.",
                SigmaEpsilonPerformanceWarning,
            )

        shp_source_inverse = generalized_inverse(shp_source)

    if not isinstance(values_source, ndarray):
        values_source = np.array(values_source)

    multi_dimensional = len(values_source.shape) > 1
    shp_target = shp_fnc(x_target)  # (nP_target, nNE)

    if len(values_source.shape) > 1 and axis is None:
        axis = -1

    if isinstance(axis, int):
        if not multi_dimensional:
            raise ValueError(
                "If 'axis' is provided, 'values_source' must be multidimensional."
            )

    if multi_dimensional:
        values_source = np.moveaxis(values_source, axis, -1)
        *array_axes, nP_source = values_source.shape
        nX = np.prod(array_axes)
        values_source = np.reshape(values_source, (nX, nP_source))
        nP_target = shp_target.shape[0]
        result = np.zeros((nX, nP_target))
        # (nP_T x nNE) @ (nNE x nP_S) @ (nX, nP_S) -> (nX, nP_T)
        _approximate_multi(shp_target @ shp_source_inverse, values_source, result)
        result = np.reshape(result, tuple(array_axes) + (nP_target,))
        result = np.moveaxis(result, -1, axis)
        values_source = np.moveaxis(values_source, -1, axis)
    else:
        # (nP_T x nNE) @ (nNE x nP_S) @ (nNE)
        result = shp_target @ shp_source_inverse @ values_source

    return result


class LagrangianCellApproximator:
    """
    An approximator for Lagrangian cells. It can be constructed directly or using
    a cell class from the library.

    Parameters
    ----------
    cell_class: :class:`~sigmaepsilon.mesh.data.polycell.PolyCell`
        A Lagrangian cell class that provides the batteries for interpolation.
        The capabilities of this class determines the nature and accuracy of the
        interpolation/extrapolation.
    x_source: Iterable, Optional
        The process of interpolation involves calculating the inverse of a matrix.
        If you plan to use an interpolator many times using the same set of source points,
        it is a good idea to fed the instance with these coordinates at the time of
        instantiation. This way the expensive part of the calculation is only done once,
        and subsequent evaluations are faster. Default is None.

    Notes
    -----
    Depending on the number of nodes of the element (hence the order of the approximation
    functions), the approximation may be exact interpolation or some kind of regression.
    For instance, if you try to extrapolate from 3 values using a 2-noded line element,
    the approximator is overfitted and the approximation is an ecaxt one only if all the
    data values fit a line perfectly.

    Examples
    --------
    Create an approximator using 8-noded hexahedrons.

    >>> from sigmaepsilon.mesh import LagrangianCellApproximator
    >>> from sigmaepsilon.mesh.cells import H8
    >>> approximator = LagrangianCellApproximator(H8)

    The data to feed the approximator:

    >>> source_coordinates = H8.Geometry.master_coordinates() / 2
    >>> source_values = [1, 2, 3, 4, 5, 6, 7, 8]
    >>> target_coordinates = H8.Geometry.master_coordinates() * 2

    The desired data at the target locations:

    >>> target_values = approximator(
    ...     source=source_coordinates,
    ...     target=target_coordinates,
    ...     values=source_values
    ... )

    This approximator can also be created using the class diretly:

    >>> approximator = H8.Geometry.approximator()

    If you want to reuse the approximator with the same set of source coordinates
    many times, you can feed these points to the approximator at instance creation:

    >>> approximator = H8.Geometry.approximator(source_coordinates)
    >>> approximator = LagrangianCellApproximator(H8, source_coordinates)

    Then, only source values and target coordinates have to be provided for
    approximation to happen (in fact, you will get an Exception of you provide
    source coordinates both at creation and approximator):

    >>> target_values = approximator(
    ...     target=target_coordinates,
    ...     values=source_values
    ... )

    To approximator multidimensional data, you have to carefully organize the
    input values for utmost performance. The memory layout is optimal if the axis
    that goes along the input points is the last one:

    >>> approximator = H8.Geometry.approximator()

    >>> source_values = np.random.rand(10, 2, 8)
    >>> approximator(
    ...     source=source_coordinates,
    ...     values=source_values,
    ...     target=target_coordinates[:3]
    ... ).shape
    (10, 2, 3)

    If it is not the last axis, you can use the 'axis' parameter:

    >>> source_values = np.random.rand(8, 2, 10)
    >>> approximator(
    ...     source=source_coordinates,
    ...     values=source_values,
    ...     target=target_coordinates[:3],
    ...     axis=0,
    ... ).shape
    (3, 2, 10)
    """

    approximator_function: Callable = _approximator

    def __init__(self, cell_class: Any, x_source: Iterable = None):
        shp_fnc: Callable = _get_shape_function_evaluator(cell_class)

        if isinstance(x_source, Iterable):
            shp_source = shp_fnc(np.array(x_source))  # (nP_source, nNE)
            self._source_shp_inverse = generalized_inverse(shp_source)
        else:
            self._source_shp_inverse = None

        self._approximator = partial(
            self.__class__.approximator_function,
            cell_class,
            shp_source_inverse=self._source_shp_inverse,
        )

    def __call__(
        self,
        *,
        values: Iterable,
        target: Iterable,
        source: Iterable = None,
        axis: int = None,
    ) -> ndarray:
        if source is not None and self._source_shp_inverse is not None:
            raise Exception("The approximator is already fed with source coordinates.")

        return self._approximator(
            x_source=source,
            x_target=target,
            values_source=values,
            axis=axis,
            shp_source_inverse=self._source_shp_inverse,
        )



from typing import Tuple, Union, Iterable

import numpy as np
from numpy import ndarray
from numba import njit, prange

from .utils.topology import unique_topo_data, detach_mesh_bulk, transform_topology
from .utils import (
    center_of_points,
    k_nearest_neighbours as knn,
    knn_to_lines,
    coords_to_3d,
)

__cache = True


__all__ = ["grid", "gridQ4", "gridQ8", "gridQ9", "gridH8", "gridH27", "knngridL2"]


def grid(
    size: Tuple[float] = None,
    shape: Union[int, Tuple[int]] = None,
    eshape: Union[str, Tuple[int]] = None,
    shift: Iterable = None,
    start: int = 0,
    bins: Iterable = None,
    centralize: bool = False,
    path: Iterable = None,
    **kwargs,
) -> Tuple[ndarray, ndarray]:
    """
    Crates a 1d, 2d or 3d grid for different patterns and returnes the raw data.
    If you want a more high level mesh object, consider using the :class:`~sigmaepsilon.mesh.grid.Grid`
    class, which calls this method to generate a :class:`~sigmaepsilon.mesh.data.polydata.PolyData` instance.

    Parameters
    ----------
    size: Tuple[float], Optional
        A 2-tuple, containg side lengths of a rectangular domain.
        Should be provided alongside `shape`. Default is None.
    shape: Union[int, Tuple[int]], Optional
        An integer or a tuple of integers, describing number of cells along
        coordinate directions. Should be provided alongside `size`. Default is None.
    eshape: str or Tuple, Optional
        This specifies element shape.
        Supported strings are thw following:
          - 'Q4' : 4-noded quadrilateral
          - 'Q9' : 9-noded quadrilateral
          - 'H8' : 8-noded hexagon
          - 'H27' : 27-noded hexagon
    shift: numpy.ndarray, Optional
        1d float array, specifying a translation.
    start: int, Optional
        Starting value for node numbering. Default is 0.
    bins: numpy.ndarray, Optional
        Numpy array or an iterable of numy arrays.
    centralize: bool, Optional
        If True, the returned coordinates are centralized.
    path: Iterable, Optional
        A 1d iterable used to rewire the nodes of the resulting grid. If provided, the
        j-th node of the i-th cell becomes ``topo[i, path[j]]``. Default is None.

    Notes
    -----
    1) The returned topology may not be compliant with vtk. If you want to use
    the results of this call to build a vtk model, you have to account for this
    (for instance through the 'path' parameter). Optionally, you can use the dedicated
    grid generation routines of this module.
    2) If you'd rather get the result as a `PolyData`, use the `Grid` class.

    Returns
    -------
    numpy.ndarray
        A numpy float array of coordinates.
    numpy.ndarray
        A numpy integer array describing the topology.

    Examples
    --------
    Create a simple hexagonal mesh of 8 x 6 x 2 cells

    >>> from sigmaepsilon.mesh import grid
    >>> size = 80, 60, 20
    >>> shape = 8, 6, 2
    >>> mesh = (coords, topo) = grid(size=size, shape=shape, eshape='H8')

    Create a mesh of 4-noded quadrilaterals

    >>> gridparams = {
    ...     'size' : (1200, 600),
    ...     'shape' : (30, 15),
    ...     'eshape' : (2, 2),
    ...     'start' : 0
    ... }
    >>> coordsQ4, topoQ4 = grid(**gridparams)

    The same mesh with 6-noded quadrialterals, 2 in x direction, 3 in y direction

    >>> gridparams['eshape'] = (2, 3)
    >>> coordsQ4, topoQ4 = grid(**gridparams)

    See also
    --------
    :class:`~sigmaepsilon.mesh.grid.Grid`
    :func:`~sigmaepsilon.mesh.grid.gridQ4`
    :func:`~sigmaepsilon.mesh.grid.gridQ9`
    :func:`~sigmaepsilon.mesh.grid.gridH8`
    :func:`~sigmaepsilon.mesh.grid.gridH27`
    """
    if size is not None:
        nDime = len(size)
    elif bins is not None:
        nDime = len(bins)
    elif shape is not None:
        nDime = len(shape)

    if shift is None:
        shift = np.zeros(nDime)

    if eshape is None:
        eshape = np.full(nDime, 2, dtype=int)
    elif isinstance(eshape, str):
        if eshape == "Q4":
            return gridQ4(
                size=size,
                shape=shape,
                shift=shift,
                start=start,
                centralize=centralize,
                bins=bins,
                **kwargs,
            )
        elif eshape == "Q8":
            return gridQ8(
                size=size,
                shape=shape,
                shift=shift,
                start=start,
                centralize=centralize,
                bins=bins,
                **kwargs,
            )
        elif eshape == "Q9":
            return gridQ9(
                size=size,
                shape=shape,
                shift=shift,
                start=start,
                centralize=centralize,
                bins=bins,
                **kwargs,
            )
        elif eshape == "H8":
            return gridH8(
                size=size,
                shape=shape,
                shift=shift,
                start=start,
                centralize=centralize,
                bins=bins,
                **kwargs,
            )
        elif eshape == "H27":
            return gridH27(
                size=size,
                shape=shape,
                shift=shift,
                start=start,
                centralize=centralize,
                bins=bins,
                **kwargs,
            )
        else:
            raise NotImplementedError

    if bins is not None:
        if len(bins) == 2:
            coords, topo = grid_2d_bins(bins[0], bins[1], eshape, shift, start)
        elif len(bins) == 3:
            coords, topo = grid_3d_bins(bins[0], bins[1], bins[2], eshape, shift, start)
        else:
            raise NotImplementedError
    else:
        assert isinstance(eshape, tuple)
        if shape is None:
            shape = np.full(nDime, 1, dtype=int)
        coords, topo = rgridMT(size, shape, eshape, shift, start)

    if centralize:
        center = center_of_points(coords)
        coords[:, 0] -= center[0]
        coords[:, 1] -= center[1]
        if center.shape[0] > 2:
            coords[:, 2] -= center[2]

    if path is not None:
        if not isinstance(path, Iterable):
            raise TypeError("Argument 'path' must be an Iterable!")
        path = np.array(path, dtype=int)
        topo = transform_topology(topo, path)

    return coords_to_3d(coords), topo


def gridQ4(*args, **kwargs) -> Tuple[ndarray, ndarray]:
    """
    Customized version of :func:`grid` dedicated to Q4 elements.
    It returns a topology with vtk-compliant node numbering.

    In terms of parameters, this function have to be called the
    same way `grid` would be called, except the parameter
    `eshape` being obsolete.

    Example
    -------
    Creating a mesh of 4-noded quadrilaterals

    >>> from sigmaepsilon.mesh.grid import gridQ4
    >>> gridparams = {
    ...     'size' : (1200, 600),
    ...     'shape' : (30, 15),
    ...     'origo' : (0, 0),
    ...     'start' : 0
    ... }
    >>> coordsQ4, topoQ4 = gridQ4(**gridparams)
    """
    coords, topo = grid(*args, eshape=(2, 2), **kwargs)
    path = np.array([0, 2, 3, 1], dtype=int)
    return coords_to_3d(coords), transform_topology(topo, path)


def gridQ9(*args, **kwargs) -> Tuple[ndarray, ndarray]:
    """
    Customized version of `grid` dedicated to Q9 elements.
    It returns a topology with vtk-compliant node numbering.

    In terms of parameters, this function have to be called the
    same way `grid` would be called, except the parameter
    `eshape` being obsolete.
    """
    coords, topo = grid(*args, eshape=(3, 3), **kwargs)
    path = np.array([0, 6, 8, 2, 3, 7, 5, 1, 4], dtype=int)
    return coords_to_3d(coords), transform_topology(topo, path)


def gridQ8(*args, **kwargs) -> Tuple[ndarray, ndarray]:
    """
    Customized version of `grid` dedicated to Q8 elements.
    It returns a topology with vtk-compliant node numbering.

    In terms of parameters, this function have to be called the
    same way `grid` would be called, except the parameter
    `eshape` being obsolete.
    """
    coords, topo = gridQ9(*args, **kwargs)
    return detach_mesh_bulk(coords, topo[:, :-1])


def gridH8(*args, **kwargs) -> Tuple[ndarray, ndarray]:
    """
    Customized version of `grid` dedicated to H8 elements.
    It returns a topology with vtk-compliant node numbering.

    In terms of parameters, this function have to be called the
    same way `grid` would be called, except the parameter
    `eshape` being obsolete.
    """
    coords, topo = grid(*args, eshape=(2, 2, 2), **kwargs)
    path = np.array([0, 4, 6, 2, 1, 5, 7, 3], dtype=int)
    return coords_to_3d(coords), transform_topology(topo, path)


# fmt: off
def gridH27(*args, **kwargs) -> Tuple[ndarray, ndarray]:
    """
    Customized version of `grid` dedicated to H27 elements.
    It returns a topology with vtk-compliant node numbering.

    In terms of parameters, this function have to be called the
    same way `grid` would be called, except the parameter
    `eshape` being obsolete.
    """
    coords, topo = grid(*args, eshape=(3, 3, 3), **kwargs)
    path = np.array(
        [
            0, 18, 24, 6, 2, 20, 26, 8, 9, 21,
            15, 3, 11, 23, 17, 5, 1, 19, 25, 7,
            4, 22, 10, 16, 12, 14, 13,
        ],
        dtype=int,
    )
    return coords, transform_topology(topo, path)
# fmt: on


@njit(nogil=True, cache=__cache)
def _rgridST(size, shape, eshape, shift, start=0):
    """
    Legacy code for single-thread implementation, for educational purpose only.
    Use rgridMT.
    """
    nDime = len(size)

    if nDime == 1:
        lX = size
        ndivX = shape
        nNodeX = eshape
        nX = ndivX * (nNodeX - 1) + 1
        dX = lX / ndivX
        ddX = dX / (nNodeX - 1)
        numCell = ndivX
        numPoin = nX
        numNode = nNodeX
    elif nDime == 2:
        lX, lY = size
        ndivX, ndivY = shape
        nNodeX, nNodeY = eshape
        nX = ndivX * (nNodeX - 1) + 1
        nY = ndivY * (nNodeY - 1) + 1
        dX = lX / ndivX
        dY = lY / ndivY
        ddX = dX / (nNodeX - 1)
        ddY = dY / (nNodeY - 1)
        numCell = ndivX * ndivY
        numPoin = nX * nY
        numNode = nNodeX * nNodeY
    elif nDime == 3:
        lX, lY, lZ = size
        ndivX, ndivY, ndivZ = shape
        nNodeX, nNodeY, nNodeZ = eshape
        nX = ndivX * (nNodeX - 1) + 1
        nY = ndivY * (nNodeY - 1) + 1
        nZ = ndivZ * (nNodeZ - 1) + 1
        dX = lX / ndivX
        dY = lY / ndivY
        dZ = lZ / ndivZ
        ddX = dX / (nNodeX - 1)
        ddY = dY / (nNodeY - 1)
        ddZ = dZ / (nNodeZ - 1)
        numCell = ndivY * ndivX * ndivZ
        numPoin = nX * nY * nZ
        numNode = nNodeX * nNodeY * nNodeZ

    # set up nodal coordinates
    coords = np.zeros((numPoin, nDime))
    topo = np.zeros((numCell, numNode), dtype=np.int64)
    elem = -1
    if nDime == 1:
        for i in range(1, ndivX + 1):
            elem += 1
            ne = -1
            n = (nNodeX - 1) * (i - 1)
            for j in range(1, nNodeX + 1):
                n += 1
                coords[n - 1, 0] = shift + dX * (i - 1) + ddX * (j - 1)
                ne += 1
                topo[elem, ne] = n
    elif nDime == 2:
        for i in range(1, ndivX + 1):
            for j in range(1, ndivY + 1):
                elem += 1
                ne = -1
                for k in range(1, nNodeX + 1):
                    for m in range(1, nNodeY + 1):
                        n = (
                            ((nNodeX - 1) * (i - 1) + k - 1) * nY
                            + (nNodeY - 1) * (j - 1)
                            + m
                        )
                        coords[n - 1, 0] = shift[0] + dX * (i - 1) + ddX * (k - 1)
                        coords[n - 1, 1] = shift[1] + dY * (j - 1) + ddY * (m - 1)
                        ne += 1
                        topo[elem, ne] = n
    elif nDime == 3:
        for i in range(1, ndivX + 1):
            for j in range(1, ndivY + 1):
                for k in range(1, ndivZ + 1):
                    elem += 1
                    ne = -1
                    for m in range(1, nNodeX + 1):
                        for q in range(1, nNodeY + 1):
                            for p in range(1, nNodeZ + 1):
                                n = (
                                    (
                                        ((nNodeX - 1) * (i - 1) + m - 1) * nY
                                        + (nNodeY - 1) * (j - 1)
                                        + q
                                        - 1
                                    )
                                    * nZ
                                    + (nNodeZ - 1) * (k - 1)
                                    + p
                                )
                                coords[n - 1, 0] = (
                                    shift[0] + dX * (i - 1) + ddX * (m - 1)
                                )
                                coords[n - 1, 1] = (
                                    shift[1] + dY * (j - 1) + ddY * (q - 1)
                                )
                                coords[n - 1, 2] = (
                                    shift[2] + dZ * (k - 1) + ddZ * (p - 1)
                                )
                                ne += 1
                                topo[elem, ne] = n
    start -= 1
    return coords, topo + start


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def rgridMT(size, shape, eshape, shift, start=0):
    nDime = len(size)
    if nDime == 1:
        lX = size[0]
        ndivX = shape[0]
        nNodeX = eshape[0]
        nX = ndivX * (nNodeX - 1) + 1
        dX = lX / ndivX
        ddX = dX / (nNodeX - 1)
        numCell = ndivX
        numPoin = nX
        numNode = nNodeX
    elif nDime == 2:
        lX, lY = size
        ndivX, ndivY = shape
        nNodeX, nNodeY = eshape
        nX = ndivX * (nNodeX - 1) + 1
        nY = ndivY * (nNodeY - 1) + 1
        dX = lX / ndivX
        dY = lY / ndivY
        ddX = dX / (nNodeX - 1)
        ddY = dY / (nNodeY - 1)
        numCell = ndivX * ndivY
        numPoin = nX * nY
        numNode = nNodeX * nNodeY
    elif nDime == 3:
        lX, lY, lZ = size
        ndivX, ndivY, ndivZ = shape
        nNodeX, nNodeY, nNodeZ = eshape
        nX = ndivX * (nNodeX - 1) + 1
        nY = ndivY * (nNodeY - 1) + 1
        nZ = ndivZ * (nNodeZ - 1) + 1
        dX = lX / ndivX
        dY = lY / ndivY
        dZ = lZ / ndivZ
        ddX = dX / (nNodeX - 1)
        ddY = dY / (nNodeY - 1)
        ddZ = dZ / (nNodeZ - 1)
        numCell = ndivY * ndivX * ndivZ
        numPoin = nX * nY * nZ
        numNode = nNodeX * nNodeY * nNodeZ
    # set up nodal coordinates
    coords = np.zeros((numPoin, nDime))
    topo = np.zeros((numCell, numNode), dtype=np.int64)
    elem = -1
    if nDime == 1:
        for i in prange(1, ndivX + 1):
            elem = i - 1
            for j in prange(1, nNodeX + 1):
                n = (nNodeX - 1) * (i - 1) + j
                ne = j - 1
                coords[n - 1, 0] = shift[0] + dX * (i - 1) + ddX * (j - 1)
                topo[elem, ne] = n
    elif nDime == 2:
        for i in prange(1, ndivX + 1):
            for j in prange(1, ndivY + 1):
                elem = (i - 1) * ndivY + j - 1
                for k in prange(1, nNodeX + 1):
                    for m in prange(1, nNodeY + 1):
                        n = (
                            ((nNodeX - 1) * (i - 1) + k - 1) * nY
                            + (nNodeY - 1) * (j - 1)
                            + m
                        )
                        ne = (k - 1) * nNodeY + m - 1
                        coords[n - 1, 0] = shift[0] + dX * (i - 1) + ddX * (k - 1)
                        coords[n - 1, 1] = shift[1] + dY * (j - 1) + ddY * (m - 1)
                        topo[elem, ne] = n
    elif nDime == 3:
        for i in prange(1, ndivX + 1):
            for j in prange(1, ndivY + 1):
                for k in prange(1, ndivZ + 1):
                    elem = (i - 1) * ndivY * ndivZ + (j - 1) * ndivZ + k - 1
                    for m in prange(1, nNodeX + 1):
                        for q in prange(1, nNodeY + 1):
                            for p in prange(1, nNodeZ + 1):
                                n = (
                                    (
                                        ((nNodeX - 1) * (i - 1) + m - 1) * nY
                                        + (nNodeY - 1) * (j - 1)
                                        + q
                                        - 1
                                    )
                                    * nZ
                                    + (nNodeZ - 1) * (k - 1)
                                    + p
                                )
                                ne = (
                                    (m - 1) * nNodeY * nNodeZ + (q - 1) * nNodeZ + p - 1
                                )
                                coords[n - 1, 0] = (
                                    shift[0] + dX * (i - 1) + ddX * (m - 1)
                                )
                                coords[n - 1, 1] = (
                                    shift[1] + dY * (j - 1) + ddY * (q - 1)
                                )
                                coords[n - 1, 2] = (
                                    shift[2] + dZ * (k - 1) + ddZ * (p - 1)
                                )
                                topo[elem, ne] = n
    start -= 1
    return coords, topo + start


@njit(nogil=True, parallel=True, cache=__cache)
def grid_2d_bins(xbins, ybins, eshape, shift, start=0):
    # shape of cells
    nEX = len(xbins) - 1
    nEY = len(ybins) - 1
    # number of cells
    nC = nEX * nEY
    # number of nodes
    nNEX, nNEY = eshape
    nDIVX = nNEX - 1
    nDIVY = nNEY - 1
    nNE = nNEX * nNEY
    nNX = nEX * nDIVX + 1
    nNY = nEY * nDIVY + 1
    nN = nNX * nNY
    # create grid
    coords = np.zeros((nN, 2))
    topo = np.zeros((nC, nNE), dtype=np.int64)
    for i in prange(nEX):
        ddX = (xbins[i + 1] - xbins[i]) / nDIVX
        for j in prange(nEY):
            ddY = (ybins[j + 1] - ybins[j]) / nDIVY
            iE = i * nEY + j
            for p in prange(nNEX):
                for q in prange(nNEY):
                    n = (nDIVX * i + p) * nNY + nDIVY * j + q
                    coords[n, 0] = shift[0] + xbins[i] + ddX * p
                    coords[n, 1] = shift[1] + ybins[j] + ddY * q
                    iNE = p * nNEY + q
                    topo[iE, iNE] = n
    return coords, topo + start


@njit(nogil=True, parallel=True, cache=__cache)
def grid_3d_bins(xbins, ybins, zbins, eshape, shift, start=0):
    # size
    lX = xbins.max() - xbins.min()
    lY = ybins.max() - ybins.min()
    lZ = zbins.max() - zbins.min()
    # shape of cells
    nEX = len(xbins) - 1
    nEY = len(ybins) - 1
    nEZ = len(zbins) - 1
    # number of cells
    nC = nEX * nEY * nEZ
    # number of nodes
    nNEX, nNEY, nNEZ = eshape
    nDIVX = nNEX - 1
    nDIVY = nNEY - 1
    nDIVZ = nNEZ - 1
    nNE = nNEX * nNEY * nNEZ
    nNX = nEX * nDIVX + 1
    nNY = nEY * nDIVY + 1
    nNZ = nEZ * nDIVZ + 1
    nN = nNX * nNY * nNZ
    # create grid
    coords = np.zeros((nN, 3), dtype=xbins.dtype)
    topo = np.zeros((nC, nNE), dtype=np.int64)
    for i in prange(nEX):
        ddX = (xbins[i + 1] - xbins[i]) / nDIVX
        for j in prange(nEY):
            ddY = (ybins[j + 1] - ybins[j]) / nDIVY
            for k in prange(nEZ):
                ddZ = (zbins[k + 1] - zbins[k]) / nDIVZ
                iE = i * nEZ * nEY + j * nEZ + k
                for p in prange(nNEX):
                    for q in prange(nNEY):
                        for r in prange(nNEZ):
                            n = (
                                ((nDIVX * i + p) * nNY + nDIVY * j + q) * nNZ
                                + nDIVZ * k
                                + r
                            )
                            coords[n, 0] = shift[0] + xbins[i] + ddX * p
                            coords[n, 1] = shift[1] + ybins[j] + ddY * q
                            coords[n, 2] = shift[2] + zbins[k] + ddZ * r
                            iNE = p * nNEZ * nNEY + q * nNEZ + r
                            topo[iE, iNE] = n
    return coords, topo + start


def knngridL2(
    *args, max_distance: float = None, k: int = 3, X: ndarray = None, **kwargs
) -> Tuple[ndarray, ndarray]:
    """
    Returns a KNN grid of L2 lines. First a grid of points is created
    using :func:``grid``, then points are connected based on a KNN-tree.

    Parameters
    ----------
    *args: tuple, Optional
        Positional arguments forwarded to :func:``grid``.
    max_distance: float, Optional
        Maximum distance allowed. Default is None.
    k: int, Optional
        Number of neighbours for a given point.
    X: numpy.ndarray, Optional
        Coordinates of a pointcloud. If provided, `args` and `kwargs`
        are ignored. Default is None.
    **kwargs: dict, Optional
        Keyword arguments forwarded to :func:``grid``.

    See also
    --------
    :func:`~sigmaepsilon.mesh.utils.knn.k_nearest_neighbours`
    :func:`~sigmaepsilon.mesh.utils.knn.knn_to_lines`
    """
    if X is None:
        X, _ = grid(*args, **kwargs)
    i = knn(X, X, k=k, max_distance=max_distance)
    T, _ = unique_topo_data(knn_to_lines(i))
    di = T[:, 0] - T[:, -1]
    inds = np.where(di != 0)[0]
    return detach_mesh_bulk(X, T[inds, :])



from typing import Tuple, List

import numpy as np
from numpy import ndarray
from sympy import symbols

from ..geometry import PolyCellGeometry2d
from ..data.polycell import PolyCell
from ..utils.cells.q9 import (
    shp_Q9_multi,
    dshp_Q9_multi,
    shape_function_matrix_Q9_multi,
    monoms_Q9,
)
from ..utils.numint import Gauss_Legendre_Quad_9
from ..utils.topology import Q4_to_T3, Q9_to_Q4


class Q9(PolyCell):
    """
    Polygon class for 9-noded biquadratic quadrilaterals.
    """

    label = "Q9"

    class Geometry(PolyCellGeometry2d):
        number_of_nodes = 9
        vtk_cell_id = 28
        shape_function_evaluator = shp_Q9_multi
        shape_function_matrix_evaluator = shape_function_matrix_Q9_multi
        shape_function_derivative_evaluator = dshp_Q9_multi
        monomial_evaluator = monoms_Q9
        quadrature = {
            "full": Gauss_Legendre_Quad_9,
            "geometry": "full",
        }

        @classmethod
        def polybase(cls) -> Tuple[List]:
            """
            Retruns the polynomial base of the master element.

            Returns
            -------
            list
                A list of SymPy symbols.
            list
                A list of monomials.
            """
            locvars = r, s = symbols("r, s", real=True)
            monoms = [
                1,
                r,
                s,
                r * s,
                r**2,
                s**2,
                r * s**2,
                s * r**2,
                s**2 * r**2,
            ]
            return locvars, monoms

        @classmethod
        def master_coordinates(cls) -> ndarray:
            """
            Returns local coordinates of the cell.

            Returns
            -------
            numpy.ndarray
            """
            return np.array(
                [
                    [-1.0, -1.0],
                    [1.0, -1.0],
                    [1.0, 1.0],
                    [-1.0, 1.0],
                    [0.0, -1.0],
                    [1.0, 0.0],
                    [0.0, 1.0],
                    [-1.0, 0.0],
                    [0.0, 0.0],
                ]
            )

        @classmethod
        def master_center(cls) -> ndarray:
            """
            Returns the local coordinates of the center of the cell.

            Returns
            -------
            numpy.ndarray
            """
            return np.array([0.0, 0.0])

        @classmethod
        def trimap(cls) -> ndarray:
            return np.array(
                [
                    [0, 4, 8],
                    [0, 8, 7],
                    [4, 1, 5],
                    [4, 5, 8],
                    [8, 5, 2],
                    [8, 2, 6],
                    [7, 8, 6],
                    [7, 6, 3],
                ],
                dtype=int,
            )

    def to_triangles(self) -> ndarray:
        return Q4_to_T3(*Q9_to_Q4(None, self.topology().to_numpy()))[1]



import warnings

import numpy as np
from meshio import Mesh as MeshioMesh

from sigmaepsilon.core.warning import SigmaEpsilonWarning

from ..data import PolyData, PointData, PolyCell
from ..space import CartesianFrame

from ..helpers import meshio_to_celltype, importers
from ..utils.space import frames_of_surfaces, frames_of_lines

__all__ = ["from_meshio"]


def from_meshio(mesh: MeshioMesh) -> PolyData:
    """
    Returns a :class:`~sigmaepsilon.mesh.polydata.PolyData` instance from a :class:`meshio.Mesh` instance.

    .. note::
        See https://github.com/nschloe/meshio for formats supported by
        ``meshio``. Be sure to install ``meshio`` with ``pip install
        meshio`` if you wish to use it.
    """
    GlobalFrame = CartesianFrame(dim=3)

    coords = mesh.points
    pd = PointData(coords=coords, frame=GlobalFrame)
    polydata = PolyData(pd)

    for cb in mesh.cells:
        cd = None
        cbtype = cb.type
        celltype: PolyCell = meshio_to_celltype.get(cbtype, None)
        if celltype:
            topo = np.array(cb.data, dtype=int)

            NDIM = celltype.Geometry.number_of_spatial_dimensions
            if NDIM == 1:
                frames = frames_of_lines(coords, topo)
            elif NDIM == 2:
                frames = frames_of_surfaces(coords, topo)
            elif NDIM == 3:
                frames = GlobalFrame

            cd = celltype(topo=topo, frames=frames)
            polydata[cbtype] = PolyData(cd)
        else:
            if cbtype != "vertex":  # pragma: no cover
                warnings.warn(
                    f"Cells of type '{cbtype}' are not supported here.",
                    SigmaEpsilonWarning,
                )

    return polydata


importers["meshio"] = from_meshio



# -*- coding: utf-8 -*-
from enum import IntEnum, unique

_meshio_to_vtk = {"triangle": "VTK_TRIANGLE"}


@unique
class vtkCellTypes(IntEnum):
    # # Linear cells
    # VTK_EMPTY_CELL = 0
    # VTK_VERTEX = 1
    # VTK_POLY_VERTEX = 2
    VTK_LINE = 3
    # VTK_POLY_LINE = 4
    VTK_TRIANGLE = 5
    # VTK_TRIANGLE_STRIP = 6
    # VTK_POLYGON = 7
    # VTK_PIXEL = 8
    VTK_QUAD = 9
    VTK_TETRA = 10
    # VTK_VOXEL = 11
    VTK_HEXAHEDRON = 12
    VTK_WEDGE = 13
    # VTK_PYRAMID = 14
    # VTK_PENTAGONAL_PRISM = 15
    # VTK_HEXAGONAL_PRISM = 16

    # # Quadratic, isoparametric cells
    VTK_QUADRATIC_EDGE = 21
    VTK_QUADRATIC_TRIANGLE = 22
    VTK_QUADRATIC_QUAD = 23
    # VTK_QUADRATIC_POLYGON = 36
    VTK_QUADRATIC_TETRA = 24
    # VTK_QUADRATIC_HEXAHEDRON = 25
    # VTK_QUADRATIC_WEDGE = 26
    # VTK_QUADRATIC_PYRAMID = 27
    VTK_BIQUADRATIC_QUAD = 28
    VTK_TRIQUADRATIC_HEXAHEDRON = 29
    # VTK_QUADRATIC_LINEAR_QUAD = 30
    # VTK_QUADRATIC_LINEAR_WEDGE = 31
    VTK_BIQUADRATIC_QUADRATIC_WEDGE = 32
    # VTK_BIQUADRATIC_QUADRATIC_HEXAHEDRON = 33
    # VTK_BIQUADRATIC_TRIANGLE = 34

    # # Cubic, isoparametric cell
    # VTK_CUBIC_LINE = 35

    # # Special class of cells formed by convex group of points
    # VTK_CONVEX_POINT_SET = 41

    # # Polyhedron cell (consisting of polygonal faces)
    # VTK_POLYHEDRON = 42

    # # Higher order cells in parametric form
    # VTK_PARAMETRIC_CURVE = 51
    # VTK_PARAMETRIC_SURFACE = 52
    # VTK_PARAMETRIC_TRI_SURFACE = 53
    # VTK_PARAMETRIC_QUAD_SURFACE = 54
    # VTK_PARAMETRIC_TETRA_REGION = 55
    # VTK_PARAMETRIC_HEX_REGION = 56

    # # Higher order cells
    # VTK_HIGHER_ORDER_EDGE = 60
    # VTK_HIGHER_ORDER_TRIANGLE = 61
    # VTK_HIGHER_ORDER_QUAD = 62
    # VTK_HIGHER_ORDER_POLYGON = 63
    # VTK_HIGHER_ORDER_TETRAHEDRON = 64
    # VTK_HIGHER_ORDER_WEDGE = 65
    # VTK_HIGHER_ORDER_PYRAMID = 66
    # VTK_HIGHER_ORDER_HEXAHEDRON = 67

    # # Arbitrary order Lagrange elements (formulated separated from generic higher order cells)
    # VTK_LAGRANGE_CURVE = 68
    # VTK_LAGRANGE_TRIANGLE = 69
    # VTK_LAGRANGE_QUADRILATERAL = 70
    # VTK_LAGRANGE_TETRAHEDRON = 71
    # VTK_LAGRANGE_HEXAHEDRON = 72
    # VTK_LAGRANGE_WEDGE = 73
    # VTK_LAGRANGE_PYRAMID = 74

    # # Arbitrary order Bezier elements (formulated separated from generic higher order cells)
    # VTK_BEZIER_CURVE = 75
    # VTK_BEZIER_TRIANGLE = 76
    # VTK_BEZIER_QUADRILATERAL = 77
    # VTK_BEZIER_TETRAHEDRON = 78
    # VTK_BEZIER_HEXAHEDRON = 79
    # VTK_BEZIER_WEDGE = 80
    # VTK_BEZIER_PYRAMID = 81

    @staticmethod
    def celltype(ID):
        if isinstance(ID, int):
            celltype = celltype_by_value(ID)
        elif isinstance(ID, str):
            celltype = celltype_by_name(ID)
        return celltype.name, celltype.value


def celltype_by_name(name: str):
    return vtkCellTypes[name]


def celltype_by_value(value: int):
    return vtkCellTypes(value)


def meshio_to_vtk(category: str, default=None):
    try:
        return _meshio_to_vtk[category]
    except Exception:
        return default


def CellTypeId(celltype):
    try:
        return vtkCellTypes[celltype].value
    except Exception:
        return None



# -*- coding: utf-8 -*-
from typing import Tuple

import numpy as np
from numpy import ndarray
from numba import njit, prange, vectorize

from sigmaepsilon.math.linalg import normalize, norm

from ..utils.utils import cells_coords, cell_coords

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def triangulate_cell_coords(ecoords: ndarray, trimap: ndarray) -> ndarray:
    nE = ecoords.shape[0]
    nTE, nNTE = trimap.shape
    nT = int(nE * nTE)
    nD = ecoords.shape[-1]
    res = np.zeros((nT, nNTE, nD), dtype=ecoords.dtype)
    for iE in prange(nE):
        for iTE in prange(nTE):
            iT = iE * nTE + iTE
            for iNTE in prange(nNTE):
                res[iT, iNTE, :] = ecoords[iE, trimap[iTE, iNTE], :]
    return res


@njit(nogil=True, cache=__cache)
def monoms_tri_loc(lcoord: ndarray) -> ndarray:
    return np.array([1, lcoord[0], lcoord[1]], dtype=lcoord.dtype)


@njit(nogil=True, cache=__cache)
def monoms_tri_loc_bulk(lcoord: ndarray) -> ndarray:
    res = np.ones((lcoord.shape[0], 3), dtype=lcoord.dtype)
    res[:, 1] = lcoord[:, 0]
    res[:, 2] = lcoord[:, 1]
    return res


@njit(nogil=True, cache=__cache)
def lcoords_tri(center: ndarray = None) -> ndarray:
    """
    Returns the local coordinates (r, s) of the vertices of a triangle.

    By default, it is assumed that the origo of the (r, s) system is at
    the geometric center of the triangle, unless the coordinates of geometric
    center are provided with the argument 'center'.

    Example
    -------
    >>> import numpy as np
    >>> from sigmaepsilon.mesh.utils.tri import lcoords_tri
    >>> lcoords = lcoords_tri(np.array([1/3, 1/3]))
    """
    res = np.array([[-1 / 3, -1 / 3], [2 / 3, -1 / 3], [-1 / 3, 2 / 3]])
    if center is not None:
        res += center
    return res


@njit(nogil=True, cache=__cache)
def ncenter_tri() -> ndarray:
    """
    Returns the area coordinates of the geometric center of the
    master triangle.
    """
    return np.array([1 / 3, 1 / 3, 1 / 3])


@njit(nogil=True, cache=__cache)
def shp_tri_loc(lcoord: ndarray, center: ndarray = None) -> ndarray:
    """
    Evaluates the shape functions at the parametric coordinates (r, s).

    By default, it is assumed that the origo of the (r, s) system is at
    the geometric center of the triangle, unless the coordinates of geometric
    center are provided with the argument 'center'.

    Example
    -------
    For a master triangle with centroid at the first vertex:
    >>> import numpy as np
    >>> from sigmaepsilon.mesh.utils.tri import shp_tri_loc
    >>> A1, A2, A3 = shp_tri_loc(np.array([0.0, 0.0]), np.array([1/3, 1/3]))
    """
    r, s = lcoord
    M = np.ones((3, 3), dtype=lcoord.dtype)
    M[1:, :] = lcoords_tri(center).T
    return np.linalg.inv(M) @ np.array([1, r, s], dtype=lcoord.dtype)


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_matrix_tri_loc(
    lcoord: ndarray, nDOFN: int = 2, center: ndarray = None
) -> ndarray:
    eye = np.eye(nDOFN, dtype=lcoord.dtype)
    shp = shp_tri_loc(lcoord, center)
    res = np.zeros((nDOFN, 3 * nDOFN), dtype=lcoord.dtype)
    for i in prange(3):
        res[:, i * 3 : (i + 1) * 3] = eye * shp[i]
    return res


@njit(nogil=True, cache=__cache)
def center_tri_2d(ecoords: ndarray) -> ndarray:
    return np.array(
        [np.mean(ecoords[:, 0]), np.mean(ecoords[:, 1])], dtype=ecoords.dtype
    )


@njit(nogil=True, cache=__cache)
def center_tri_3d(ecoords: ndarray) -> ndarray:
    return np.array(
        [np.mean(ecoords[:, 0]), np.mean(ecoords[:, 1]), np.mean(ecoords[:, 2])],
        dtype=ecoords.dtype,
    )


@njit(nogil=True, cache=__cache)
def area_tri(ecoords: ndarray) -> ndarray:
    """
    Returns the the signed area of a single 3-noded triangle.

    Parameters
    ----------
    ecoords: numpy.ndarray
        Element coordinates.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.

    Returns
    -------
    float
        Returns a positive number if the vertices are listed counterclockwise,
        negative if they are listed clockwise.
    """
    A = (
        (ecoords[1, 0] * ecoords[2, 1] - ecoords[2, 0] * ecoords[1, 1])
        + (ecoords[2, 0] * ecoords[0, 1] - ecoords[0, 0] * ecoords[2, 1])
        + (ecoords[0, 0] * ecoords[1, 1] - ecoords[1, 0] * ecoords[0, 1])
    )
    return A / 2


@njit(nogil=True, cache=__cache)
def inscribed_radius(ecoords: ndarray) -> ndarray:
    """
    Returns the radius of the inscribed circle of a single triangle.

    Parameters
    ----------
    ecoords: numpy.ndarray
        2d float numpy array of element coordinates.

    Notes
    -----
    If the sides have length a, b, c, we define the semiperimeter s
    to be half their sum, so s = (a+b+c)/2. Given this, the radius is
    given using the following:

        r2 = (s - a)*(s - b)*(s - c) / s.

    This function is numba-jittable in 'nopython' mode.

    Returns
    -------
    float
    """
    a = norm(ecoords[1] - ecoords[0])
    b = norm(ecoords[2] - ecoords[1])
    c = norm(ecoords[2] - ecoords[0])
    s = (a + b + c) / 2
    return np.sqrt((s - a) * (s - b) * (s - c) / s)


@njit(nogil=True, parallel=True, cache=__cache)
def inscribed_radii(ecoords: ndarray) -> ndarray:
    """
    Returns the radii of the inscribed circle of several triangles.

    Parameters
    ----------
    ecoords: numpy.ndarray
        3d float numpy array of element coordinates for multiple cells.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.

    Returns
    -------
    numpy.ndarray
        1d numpy float array
    """
    nE = ecoords.shape[0]
    res = np.zeros(nE)
    for i in prange(nE):
        res[i] = inscribed_radius(ecoords[i])
    return res


@njit(nogil=True, parallel=False, cache=__cache)
def areas_tri(ecoords: ndarray) -> ndarray:
    """
    Returns the total sum of signed areas of several triangles.

    Parameters
    ----------
    ecoords: numpy.ndarray
        3d float numpy array of element coordinates for multiple cells.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.

    Returns
    -------
    float
        The sum of areas of all triangles.
    """
    A = 0.0
    nE = len(ecoords)
    for i in prange(nE):
        A += (
            (ecoords[i, 1, 0] * ecoords[i, 2, 1] - ecoords[i, 2, 0] * ecoords[i, 1, 1])
            + (
                ecoords[i, 2, 0] * ecoords[i, 0, 1]
                - ecoords[i, 0, 0] * ecoords[i, 2, 1]
            )
            + (
                ecoords[i, 0, 0] * ecoords[i, 1, 1]
                - ecoords[i, 1, 0] * ecoords[i, 0, 1]
            )
        )
    return A / 2


@njit(nogil=True, parallel=True, cache=__cache)
def area_tri_bulk(ecoords: ndarray) -> ndarray:
    """
    Returns the signed area of several triangles.

    Parameters
    ----------
    ecoords: numpy.ndarray
        3d float numpy array of element coordinates for multiple cells.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.

    Returns
    -------
    numpy.ndarray
        1d numpy float array
    """
    nE = len(ecoords)
    res = np.zeros(nE, dtype=ecoords.dtype)
    for i in prange(nE):
        res[i] = (
            (ecoords[i, 1, 0] * ecoords[i, 2, 1] - ecoords[i, 2, 0] * ecoords[i, 1, 1])
            + (
                ecoords[i, 2, 0] * ecoords[i, 0, 1]
                - ecoords[i, 0, 0] * ecoords[i, 2, 1]
            )
            + (
                ecoords[i, 0, 0] * ecoords[i, 1, 1]
                - ecoords[i, 1, 0] * ecoords[i, 0, 1]
            )
        )
    return res / 2


@vectorize("f8(f8, f8, f8, f8, f8, f8)", target="parallel", cache=__cache)
def area_tri_u(x1, y1, x2, y2, x3, y3) -> float:
    """
    Vectorized implementation of `area_tri_bulk`.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    return (x2 * y3 - x3 * y2 + x3 * y1 - x1 * y3 + x1 * y2 - x2 * y1) / 2


@vectorize("f8(f8, f8, f8, f8, f8, f8)", target="parallel", cache=__cache)
def area_tri_u2(x1, x2, x3, y1, y2, y3) -> float:
    """
    Another vectorized implementation of `area_tri_bulk` with a different
    order of arguments.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    return (x2 * y3 - x3 * y2 + x3 * y1 - x1 * y3 + x1 * y2 - x2 * y1) / 2


@njit(nogil=True, cache=__cache)
def loc_to_glob_tri(
    lcoord: ndarray, gcoords: ndarray, center: ndarray = None
) -> ndarray:
    """
    Transformation from local to global coordinates within a triangle.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    return gcoords.T @ shp_tri_loc(lcoord, center)


@njit(nogil=True, cache=__cache)
def glob_to_loc_tri(
    gcoord: ndarray, gcoords: ndarray, center: ndarray = None
) -> ndarray:
    """
    Transformation from global to local coordinates within a triangle.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    monoms = monoms_tri_loc_bulk(gcoords)
    coeffs = np.linalg.inv(monoms)
    shp = coeffs.T @ monoms_tri_loc(gcoord)
    return lcoords_tri(center).T @ shp


@njit(nogil=True, cache=__cache)
def glob_to_nat_tri(gcoord: ndarray, ecoords: ndarray) -> ndarray:
    """
    Transformation from global to natural coordinates within a triangle.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    x, y = gcoord[0:2]
    (x1, x2, x3), (y1, y2, y3) = ecoords[:, 0], ecoords[:, 1]
    A2 = np.abs(x2 * (y3 - y1) + x1 * (y2 - y3) + x3 * (y1 - y2))
    n1 = (x2 * y3 - x * y3 - x3 * y2 + x * y2 + x3 * y - x2 * y) / A2
    n2 = (x * y3 - x1 * y3 + x3 * y1 - x * y1 - x3 * y + x1 * y) / A2
    return np.array([n1, n2, 1 - n1 - n2], dtype=gcoord.dtype)


@njit(nogil=True, parallel=True, cache=__cache)
def _glob_to_nat_tri_bulk_(points: ndarray, ecoords: ndarray) -> ndarray:
    nE = ecoords.shape[0]
    nP = points.shape[0]
    res = np.zeros((nP, nE, 3), dtype=points.dtype)
    for i in prange(nP):
        for j in prange(nE):
            res[i, j] = glob_to_nat_tri(points[i], ecoords[j])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def __pip_tri_bulk__(nat: ndarray, tol: float = 1e-12) -> ndarray:
    nP, nE = nat.shape[:2]
    res = np.zeros((nP, nE), dtype=np.bool_)
    for i in prange(nP):
        for j in prange(nE):
            c1 = np.all(nat[i, j] > (-tol))
            c2 = np.all(nat[i, j] < (1 + tol))
            res[i, j] = c1 and c2
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _pip_tri_bulk_(points: ndarray, ecoords: ndarray, tol: float = 1e-12) -> ndarray:
    nat = _glob_to_nat_tri_bulk_(points, ecoords)
    return __pip_tri_bulk__(nat, tol)


@njit(nogil=True, parallel=True, cache=__cache)
def _glob_to_nat_tri_bulk_knn_(
    points: ndarray, ecoords: ndarray, neighbours: ndarray
) -> ndarray:
    kE = neighbours.shape[1]
    nP = points.shape[0]
    res = np.zeros((nP, kE, 3), dtype=points.dtype)
    for i in prange(nP):
        for k in prange(kE):
            res[i, k, :] = glob_to_nat_tri(points[i], ecoords[neighbours[i, k]])
    return res


@njit(nogil=True, cache=__cache)
def _pip_tri_bulk_knn_(
    points: ndarray, ecoords: ndarray, neighbours: ndarray, tol: float = 1e-12
) -> ndarray:
    nat = _glob_to_nat_tri_bulk_knn_(points, ecoords, neighbours)
    return __pip_tri_bulk__(nat, tol)


@njit(nogil=True, cache=__cache)
def nat_to_glob_tri(ncoord: ndarray, ecoords: ndarray) -> ndarray:
    """
    Transformation from natural to global coordinates within a triangle.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    return ecoords.T @ ncoord


@njit(nogil=True, cache=__cache)
def loc_to_nat_tri(lcoord: ndarray, center: ndarray = None) -> ndarray:
    """
    Transformation from local to natural coordinates within a triangle.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    return shp_tri_loc(lcoord, center)


@njit(nogil=True, cache=__cache)
def nat_to_loc_tri(
    acoord: ndarray, lcoords: ndarray = None, center: ndarray = None
) -> ndarray:
    """
    Transformation from natural to local coordinates within a triangle.

    Parameters
    ----------
    acoord: numpy.ndarray
        1d NumPy array of area coordinates of a point.
    lcoords: numpy.ndarray, Optional
        2d NumPy array of parametric coordinates (r, s) of the
        master cell of a triangle.
    center: numpy.ndarray
        The local coordinates (r, s) of the geometric center
        of the master triangle. If not provided it is assumed to
        be at (0, 0).

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    if lcoords is None:
        lcoords = lcoords_tri(center)
    return acoord.T @ lcoords


@njit(nogil=True, parallel=True, cache=__cache)
def localize_points(
    points: ndarray, triangles: ndarray, coords: ndarray
) -> Tuple[ndarray, ndarray]:
    nE = triangles.shape[0]
    nC = coords.shape[0]
    ecoords = cells_coords(points, triangles)
    res = np.full(nC, -1, dtype=triangles.dtype)
    shp = np.zeros((nC, 3), dtype=points.dtype)
    for iC in prange(nC):
        for iE in prange(nE):
            nat = glob_to_nat_tri(coords[iC], ecoords[iE])
            if np.max(nat) <= 1.0:
                res[iC] = iE
                shp[iC] = nat
                break
    return res, shp


@njit(nogil=True, parallel=True, cache=__cache)
def _get_points_inside_triangles(
    points: ndarray, topo: ndarray, coords: ndarray
) -> ndarray:
    inds, _ = localize_points(points, topo, coords)
    inds[inds > -1] = 1
    inds[inds < 0] = 0
    return inds


def get_points_inside_triangles(
    points: ndarray, topo: ndarray, coords: ndarray
) -> ndarray:
    return _get_points_inside_triangles(points, topo, coords).astype(bool)


@njit(nogil=True, parallel=True, cache=__cache)
def approx_data_to_points(
    points: ndarray,
    triangles: ndarray,
    data: ndarray,
    coords: ndarray,
    defval: float = 0.0,
) -> ndarray:
    nC = coords.shape[0]
    nD = data.shape[1]
    inds, shp = localize_points(points, triangles, coords)
    res = np.full((nC, nD), defval, dtype=data.dtype)
    for iC in prange(nC):
        i = inds[iC]
        if i > -1:
            for j in prange(nD):
                res[i, j] = np.sum(data[triangles[i], j] * shp[iC])
    return res


def offset_tri(coords: ndarray, topo: ndarray, data: ndarray) -> ndarray:
    if isinstance(data, ndarray):
        alpha = np.abs(data)
        amax = alpha.max()
        if amax > 1.0:
            alpha /= amax
        return _offset_tri_(coords, topo, alpha)
    elif isinstance(data, float):
        alpha = min(abs(data), 1.0)
        return offset_tri_uniform(coords, topo, alpha)
    else:
        raise RuntimeError


@njit(nogil=True, cache=__cache)
def offset_tri_uniform(coords: ndarray, topo: ndarray, alpha: float = 0.9) -> ndarray:
    cellcoords = cells_coords(coords, topo)
    ncenter = ncenter_tri(coords.dtype)
    eye = np.eye(3, dtype=coords.dtype)
    ncoords = ncenter + (eye - ncenter) * alpha
    nE = len(topo)
    res = np.zeros(cellcoords.shape, dtype=cellcoords.dtype)
    ncoords = ncoords.astype(cellcoords.dtype)
    for iE in prange(nE):
        res[iE, :, :] = ncoords @ cellcoords[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _offset_tri_(coords: ndarray, topo: ndarray, alpha: ndarray) -> ndarray:
    cellcoords = cells_coords(coords, topo)
    ncenter = ncenter_tri()
    dn = np.eye(3, dtype=coords.dtype) - ncenter
    nE = len(topo)
    res = np.zeros(cellcoords.shape, dtype=cellcoords.dtype)
    alpha = alpha.astype(cellcoords.dtype)
    for iE in prange(nE):
        ncoords = ncenter + dn * alpha[iE]
        res[iE, :, :] = ncoords @ cellcoords[iE]
    return res


def edges_tri(triangles: ndarray) -> ndarray:
    shp = triangles.shape
    if len(shp) == 2:
        return _edges_tri(triangles)
    elif len(shp) == 3:
        return _edges_tri_pop(triangles)
    else:
        raise NotImplementedError


@njit(nogil=True, cache=__cache)
def _edges_tri(triangles: ndarray) -> ndarray:
    nE = len(triangles)
    edges = np.zeros((nE, 3, 2), dtype=triangles.dtype)
    edges[:, 0, 0] = triangles[:, 0]
    edges[:, 0, 1] = triangles[:, 1]
    edges[:, 1, 0] = triangles[:, 1]
    edges[:, 1, 1] = triangles[:, 2]
    edges[:, 2, 0] = triangles[:, 2]
    edges[:, 2, 1] = triangles[:, 0]
    return edges


@njit(nogil=True, parallel=True, cache=__cache)
def _edges_tri_pop(triangles: ndarray) -> ndarray:
    nPop, nE, _ = triangles.shape
    res = np.zeros((nPop, nE, 3, 2), dtype=triangles.dtype)
    for i in prange(nPop):
        res[i] = _edges_tri(triangles[i])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tri_glob_to_loc(
    points: ndarray, triangles: ndarray
) -> Tuple[ndarray, ndarray, ndarray]:
    nE = triangles.shape[0]
    tr = np.zeros((nE, 3, 3), dtype=points.dtype)
    res = np.zeros((nE, 3, 2), dtype=points.dtype)
    centers = np.zeros((nE, 3), dtype=points.dtype)
    for iE in prange(nE):
        centers[iE] = center_tri_3d(cell_coords(points, triangles[iE]))
        tr[iE, 0, :] = normalize(points[triangles[iE, 1]] - points[triangles[iE, 0]])
        tr[iE, 1, :] = normalize(points[triangles[iE, 2]] - points[triangles[iE, 0]])
        tr[iE, 2, :] = np.cross(tr[iE, 0, :], tr[iE, 1, :])
        tr[iE, 1, :] = np.cross(tr[iE, 2, :], tr[iE, 0, :])
        for jN in prange(3):
            vj = points[triangles[iE, jN]] - centers[iE]
            res[iE, jN, 0] = np.dot(tr[iE, 0, :], vj)
            res[iE, jN, 1] = np.dot(tr[iE, 1, :], vj)
    return res, centers, tr


if __name__ == "__main__":
    from sigmaepsilon.mesh.triang import triangulate
    from sigmaepsilon.mesh.utils.space import frames_of_surfaces, is_planar_surface
    from sigmaepsilon.math import ascont
    from time import time

    points, triangles, triobj = triangulate(size=(800, 600), shape=(100, 100))
    tricoords = cells_coords(points, triangles)

    area0 = 800 * 600

    t0 = time()
    for i in range(10):
        area1 = np.sum(area_tri_bulk(tricoords))
    print(time() - t0)

    t0 = time()
    for i in range(10):
        area2 = areas_tri(tricoords)
    print(time() - t0)

    x1 = tricoords[:, 0, 0]
    x2 = tricoords[:, 1, 0]
    x3 = tricoords[:, 2, 0]
    y1 = tricoords[:, 0, 1]
    y2 = tricoords[:, 1, 1]
    y3 = tricoords[:, 2, 1]
    t0 = time()
    for i in range(10):
        area3 = np.sum(area_tri_u2(x1, x2, x3, y1, y2, y3))
    print(time() - t0)

    for i in range(10):
        area4 = np.sum(area_tri_u(x1, y1, x2, y2, y3, y3))
    print(time() - t0)

    tri_glob_to_loc(points, triangles)
    frames = frames_of_surfaces(points, triangles)
    normals = ascont(frames[:, 2, :])
    print(is_planar_surface(normals))



try:
    import vtk

    __hasvtk__ = True
except Exception:  # pragma: no cover
    __hasvtk__ = False

try:
    import pyvista as pv

    __haspyvista__ = True
except Exception:  # pragma: no cover
    __haspyvista__ = False

try:
    import matplotlib as mpl

    __hasmatplotlib__ = True
except Exception:  # pragma: no cover
    __hasmatplotlib__ = False

try:
    import plotly.express as px
    import plotly.graph_objects as go

    __hasplotly__ = True
except Exception:  # pragma: no cover
    __hasplotly__ = False

try:
    import networkx as nx

    __hasnx__ = True
except Exception:  # pragma: no cover
    __hasnx__ = False

try:
    import k3d

    __hask3d__ = True
except Exception:  # pragma: no cover
    __hask3d__ = False



from typing import Union, Iterable
import warnings

import numpy as np

from sigmaepsilon.math.linalg import Vector, FrameLike, CartesianFrame, ReferenceFrame


class Point(Vector):
    """
    .. deprecated:: 2.3.0
       Use :class:`~sigmaepsilon.mesh.space.pointcloud.PointCloud` instead.

    A class a to handle a single point in Euclidean space.

    It inherits :class:`Vector <sigmaepsilon.math.linalg.vector.Vector>`,
    and extends its behaviour with default frame management for domain specific applications.

    If data is provided on object creation, the class can infer an appropriate default frame,
    hence the specification of such can be omitted.

    Parameters
    ----------
    frame: Union[ReferenceFrame, numpy.ndarray, Iterable], Optional
        A suitable reference frame, or an iterable representing coordinate axes of one.
        Default is None.

    Note
    ----
    1) This is class is superseded by :class:`PointCloud <sigmaepsilon.mesh.space.pointcloud.PointCloud>`.
    2) This class does not take the origo of the supporting reference frame into consideration when
    transforming coordinates between frames.

    Examples
    --------
    >>> from sigmaepsilon.mesh.space import Point
    >>> p = Point([1., 1., 1.])
    >>> type(p.frame)
    <class 'sigmaepsilon.math.linalg.frame.CartesianFrame'>

    If we want to handle more than one point:

    >>> import math
    >>> p = Point([[1., 0., 0.], [0., 1., 0.]])
    >>> A = p.frame
    >>> B = A.orient_new('Body', [0, 0, math.pi/2], 'XYZ')
    >>> point_in_B = p.show(B)
    """

    _frame_cls_ = CartesianFrame

    def __init__(
        self,
        *args,
        frame: Union[ReferenceFrame, np.ndarray, Iterable] = None,
        id: int = None,
        gid: int = None,
        **kwargs,
    ):
        warnings.warn(
            "The Point class is deprecated and will be removed in a future version. "
            "Use a PointCloud instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if frame is None:
            if len(args) > 0:
                if isinstance(args[0], np.ndarray):
                    frame = self._frame_cls_(dim=args[0].shape[-1])
                else:
                    try:
                        arg = np.array(args[0])
                        frame = self._frame_cls_(dim=arg.shape[-1])
                    except Exception:
                        frame = None
        else:
            if not isinstance(frame, self._frame_cls_):
                if isinstance(frame, FrameLike):
                    frame = self._frame_cls_(frame.axes)
                elif isinstance(frame, np.ndarray):
                    frame = self._frame_cls_(frame)
                elif isinstance(frame, Iterable):
                    frame = self._frame_cls_(np.array(frame, dtype=float))

        if not isinstance(frame, self._frame_cls_):
            raise ValueError("Invalid frame!")

        super().__init__(*args, frame=frame, **kwargs)
        self._id = id
        self._gid = id if gid is None else gid

    @property
    def id(self) -> int:
        return self._id

    @property
    def gid(self) -> int:
        return self._gid



from numba import njit, prange
import numpy as np
from numpy import ndarray

__cache = True


@njit(nogil=True, cache=__cache)
def monoms_T6(x: ndarray) -> ndarray:
    r, s = x
    return np.array([1, r, s, r**2, s**2, r * s], dtype=x.dtype)


@njit(nogil=True, cache=__cache)
def shp_T6(pcoord: ndarray) -> ndarray:
    r, s = pcoord[0:2]
    res = np.array(
        [
            2.0 * r**2 + 4.0 * r * s - r / 3.0 + 2.0 * s**2 - s / 3.0 - 1 / 9,
            2.0 * r**2 + r / 3.0 - 1 / 9,
            2.0 * s**2 + s / 3.0 - 1 / 9,
            -4.0 * r**2 - 4.0 * r * s - 4.0 * s / 3.0 + 4 / 9,
            4.0 * r * s + 4 * r / 3 + 4.0 * s / 3.0 + 4 / 9,
            -4.0 * r * s - 4 * r / 3 - 4.0 * s**2 + 4 / 9,
        ],
        dtype=pcoord.dtype,
    )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shp_T6_multi(pcoords: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, 6), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP, :] = shp_T6(pcoords[iP])
    return res


@njit(nogil=True, parallel=False, cache=__cache)
def shape_function_matrix_T6(pcoord: ndarray, ndof: int = 2) -> ndarray:
    eye = np.eye(ndof, dtype=pcoord.dtype)
    shp = shp_T6(pcoord)
    res = np.zeros((ndof, ndof * 6), dtype=pcoord.dtype)
    for i in prange(6):
        res[:, i * ndof : (i + 1) * ndof] = eye * shp[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_matrix_T6_multi(pcoords: ndarray, ndof: int = 2) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, ndof, ndof * 6), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = shape_function_matrix_T6(pcoords[iP], ndof)
    return res


@njit(nogil=True, cache=__cache)
def dshp_T6(pcoord: ndarray) -> ndarray:
    r, s = pcoord[0:2]
    res = np.array(
        [
            [4.0 * r + 4.0 * s - 1 / 3, 4.0 * r + 4.0 * s - 1 / 3],
            [4.0 * r + 1 / 3, 0.0],
            [0.0, 4.0 * s + 1 / 3],
            [-8.0 * r - 4.0 * s, -4.0 * r - 4 / 3],
            [4.0 * s + 4 / 3, 4.0 * r + 4 / 3],
            [-4.0 * s - 4 / 3, -4.0 * r - 8.0 * s],
        ]
    )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def dshp_T6_multi(pcoords: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, 6, 2), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = dshp_T6(pcoords[iP])
    return res


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def areas_T6(ecoords: ndarray, qpos: ndarray, qweight: ndarray) -> ndarray:
    nE = len(ecoords)
    res = np.zeros(nE, dtype=ecoords.dtype)
    nP = len(qweight)
    for i in range(nP):
        dshp = dshp_T6(qpos[i])
        for iE in prange(nE):
            jac = ecoords[iE].T @ dshp
            djac = np.linalg.det(jac)
            res[iE] += qweight[i] * djac
    return res



from typing import Tuple

import numpy as np
from numpy import ndarray
from numba import njit, prange

from sigmaepsilon.math.linalg import normalize, normalize2d, norm2d
from sigmaepsilon.math import atleast2d

from .utils import center_of_points, cell_center, cell_coords
from .knn import k_nearest_neighbours

__cache = True


@njit(nogil=True, cache=__cache)
def frame_of_plane(coords: ndarray) -> Tuple[ndarray, ndarray]:
    """
    Returns the frame of a planar surface. It needs
    at least 3 pointds to work properly (len(coords>=3)).

    It takes the center, the first point and a point from
    the middle to form the coordinate axes.

    Parameters
    ----------
    coords: numpy.ndarray
        2d coordinate array

    Returns
    -------
    numpy.ndarray
        3x3 global -> local DCM matrix
    """
    tr = np.zeros((3, 3), dtype=coords.dtype)
    center = center_of_points(coords)
    tr[:, 0] = normalize(coords[0] - center)
    tr[:, 1] = normalize(coords[np.int(len(coords) / 2)] - center)
    tr[:, 2] = normalize(np.cross(tr[:, 0], tr[:, 1]))
    tr[:, 1] = np.cross(tr[:, 2], tr[:, 0])
    return center, tr


@njit(nogil=True, parallel=True, cache=__cache)
def frames_of_surfaces(coords: ndarray, topo: ndarray) -> ndarray:
    """
    Returns the coordinates of the axes forming the local
    coordinate systems of the surfaces using the first three vertices
    of the cells.

    Note
    ----
    This is only working for flat 2d cells.

    Parameters
    ----------
    coords: numpy.ndarray
        2d coordinate array
    topo: numpy.ndarray
        2d point-based topology array

    Returns
    -------
    numpy.ndarray
        3d array of 3x3 transformation matrices
    """
    nE = topo.shape[0]
    tr = np.zeros((nE, 3, 3), dtype=coords.dtype)
    for iE in prange(nE):
        tr[iE, 0, :] = normalize(coords[topo[iE, 1]] - coords[topo[iE, 0]])
        tr[iE, 1, :] = normalize(coords[topo[iE, 2]] - coords[topo[iE, 0]])
        tr[iE, 2, :] = normalize(np.cross(tr[iE, 0, :], tr[iE, 1, :]))
        tr[iE, 1, :] = np.cross(tr[iE, 2, :], tr[iE, 0, :])
    return tr


@njit(nogil=True, parallel=True, cache=__cache)
def tr_cell_glob_to_loc_bulk(
    coords: ndarray, topo: ndarray
) -> Tuple[ndarray, ndarray, ndarray]:
    """
    Returns the coordinates of the cells in their local coordinate
    system, the coordinates of their centers and the coordinates of
    the axes forming their local coordinate system. The local coordinate
    systems are located at the centers of the cells.

    Parameters
    ----------
    coords: numpy.ndarray
        2d coordinate array
    topo: numpy.ndarray
        2d point-based topology array

    Returns
    -------
    numpy.ndarray
        2d coordinate array of local coordinates
    numpy.ndarray
        2d array of cell centers
    numpy.ndarray
        3d array of 3x3 transformation matrices
    """
    nE, nNE = topo.shape
    tr = np.zeros((nE, 3, 3), dtype=coords.dtype)
    res = np.zeros((nE, nNE, 2), dtype=coords.dtype)
    centers = np.zeros((nE, 3), dtype=coords.dtype)
    for iE in prange(nE):
        centers[iE] = cell_center(cell_coords(coords, topo[iE]))
        tr[iE, 0, :] = normalize(coords[topo[iE, 1]] - coords[topo[iE, 0]])
        tr[iE, 1, :] = normalize(coords[topo[iE, nNE - 1]] - coords[topo[iE, 0]])
        tr[iE, 2, :] = normalize(np.cross(tr[iE, 0, :], tr[iE, 1, :]))
        tr[iE, 1, :] = np.cross(tr[iE, 2, :], tr[iE, 0, :])
        for jN in prange(nNE):
            vj = coords[topo[iE, jN]] - centers[iE]
            res[iE, jN, 0] = np.dot(tr[iE, 0, :], vj)
            res[iE, jN, 1] = np.dot(tr[iE, 1, :], vj)
    return res, centers, tr


@njit(nogil=True, parallel=True, cache=__cache)
def _frames_of_lines_auto(coords: ndarray, topo: ndarray) -> ndarray:
    nE = topo.shape[0]
    ijk = np.eye(3)
    tr = np.zeros((nE, 3, 3), dtype=coords.dtype)
    for iE in prange(nE):
        tr[iE, 0, :] = normalize(coords[topo[iE, -1]] - coords[topo[iE, 0]])
        _dot = ijk @ tr[iE, 0, :]
        i2 = np.argmin(np.absolute(_dot))
        _dot = np.dot(ijk[i2], tr[iE, 0, :])
        tr[iE, 2, :] = normalize(ijk[i2] - tr[iE, 0, :] * _dot)
        tr[iE, 1, :] = np.cross(tr[iE, 2, :], tr[iE, 0, :])
    return tr


@njit(nogil=True, parallel=True, cache=__cache)
def _frames_of_lines_ref(coords: ndarray, topo: ndarray, refZ: ndarray) -> ndarray:
    nE, nNE = topo.shape
    nNE -= 1
    tr = np.zeros((nE, 3, 3), dtype=coords.dtype)
    for iE in prange(nE):
        tr[iE, 0, :] = normalize(coords[topo[iE, nNE]] - coords[topo[iE, 0]])
        k = refZ[iE] - coords[topo[iE, 0]]
        tr[iE, 2, :] = normalize(k - tr[iE, 0, :] * np.dot(tr[iE, 0, :], k))
        tr[iE, 1, :] = np.cross(tr[iE, 2, :], tr[iE, 0, :])
    return tr


def frames_of_lines(coords: ndarray, topo: ndarray, refZ: ndarray = None) -> ndarray:
    """
    Returns coordinate frames of line elements defined by a coordinate array
    and a topology array. The cross-sections of the line elements are
    in the local y-z plane. The direction of the local z axis can be set by
    providing reference points in the local x-z planes of the lines. If there
    are no references provided, local z axes lean towards global z.
    Other properties are determined in a way, so that x-y-z form a
    right-handed orthonormal basis.

    Parameters
    ----------
    coords: numpy.ndarray
        2d coordinate array
    topo: numpy.ndarray
        2d point-based topology array
    refZ: numpy.ndarray, Optional
        1d or 2d float array of reference points. If it is 2d, it must
        contain values for all lines defined by `topo`.
        Default is None.

    Returns
    -------
    numpy.ndarray
        3d array of 3x3 transformation matrices
    """
    topo = atleast2d(topo)
    if isinstance(refZ, ndarray):
        if len(topo.shape) == 2 and len(refZ.shape) == 1:
            _refZ = np.zeros((topo.shape[0], 3))
            _refZ[:] = refZ
        else:
            _refZ = refZ
        return _frames_of_lines_ref(coords, topo, _refZ)
    else:
        return _frames_of_lines_auto(coords, topo)


@njit(nogil=True, parallel=True, cache=__cache)
def is_planar_surface(normals: ndarray, tol: float = 1e-8) -> bool:
    """
    Returns true if all the normals point in the same direction.
    The provided normal vectors are assumed to be normalized.

    Parameters
    ----------
    normals: numpy.ndarray
        2d float array of surface normals
    tol: float
        Floating point tolerance as maximum deviation.

    Returns
    -------
    bool
        True if the surfaces whose normal vectors are provided form
        a flat surface, False otherwise.
    """
    nE = normals.shape[0]
    diffs = np.zeros(nE, dtype=normals.dtype)
    for i in prange(1, nE):
        diffs[i] = np.abs(normals[i] @ normals[0] - 1)
    return diffs.max() <= tol


@njit(nogil=True, cache=__cache)
def distances_from_point(
    coords: ndarray, p: ndarray, normalize: bool = False
) -> ndarray:
    if normalize:
        return norm2d(normalize2d(coords - p))
    else:
        return norm2d(coords - p)


@njit(nogil=True, parallel=True, cache=__cache)
def distance_matrix(x: ndarray, y: ndarray) -> ndarray:
    N = x.shape[0]
    M = y.shape[0]
    res = np.zeros((N, M), dtype=x.dtype)
    for n in prange(N):
        for m in prange(M):
            res[n, m] = np.linalg.norm(x[n] - y[m])
    return res


def index_of_closest_point(coords: ndarray, target: ndarray) -> int:
    """
    Returs the index of the point in 'coords', being closest to
    one or more targets.

    Parameters
    ----------
    coords: numpy.ndarray
        2d float array of vertex coordinates.
    target: numpy.ndarray
        1d or 2d coordinate array of the target point(s).

    Returns
    -------
    int or Iterable[int]
        One or more indices of 'coords', for which the distance from
        one or more points described by 'target' is minimal.
    """
    if len(target.shape) == 1:
        assert (
            coords.shape[1] == target.shape[0]
        ), "The dimensions of `coords` and `target` are not compatible."
        return _index_of_closest_point(coords, target)
    else:
        return k_nearest_neighbours(coords, target)


def index_of_furthest_point(coords: ndarray, target: ndarray) -> int:
    """
    Returs the index of the point in 'coords', being furthest from
    one or more targets.

    Parameters
    ----------
    coords: numpy.ndarray
        2d float array of vertex coordinates.
    target: numpy.ndarray
        1d or 2d coordinate array of the target point(s).

    Returns
    -------
    int or Iterable[int]
        One or more indices of 'coords', for which the distance from
        one or more points described by 'target' is maximal.
    """
    if len(target.shape) == 1:
        assert (
            coords.shape[1] == target.shape[0]
        ), "The dimensions of `coords` and `target` are not compatible."
        return _index_of_furthest_point(coords, target)
    else:
        d = distance_matrix(target, coords)
        return np.argmax(d, axis=1)


@njit(nogil=True, cache=__cache)
def _index_of_closest_point(coords: ndarray, p: ndarray) -> int:
    return np.argmin(distances_from_point(coords, p))


@njit(nogil=True, cache=__cache)
def _index_of_furthest_point(coords: ndarray, p: ndarray) -> int:
    return np.argmax(distances_from_point(coords, p))


@njit(nogil=True, parallel=True, cache=__cache)
def is_line(coords: ndarray, tol=1e-8) -> bool:
    """
    Returns true if all the normals point in the same direction.
    The provided normal vectors are assumed to be normalized.

    Parameters
    ----------
    coords: numpy.ndarray
        2d float array of point coordinates
    tol: float
        Floating point tolerance as maximum deviation.

    Returns
    -------
    bool
        True if all absolute deviations from the line between the first
        and the last point is smaller than 'tol'.
    """
    nP = coords.shape[0]
    c = normalize2d(move_points(coords, -coords[0]))
    d = c[-1] - c[0]
    diffs = np.zeros(nP, dtype=coords.dtype)
    for i in prange(1, nP):
        diffs[i] = np.abs(c[i] @ d)
    return diffs.max() <= tol


@njit(nogil=True, parallel=True, cache=__cache)
def is_planar(coords: ndarray, tol: float = 1e-8) -> bool:
    """
    Returns true if all the points fit on a planar surface.

    Parameters
    ----------
    coords: numpy.ndarray
        2d float array of point coordinates
    tol: float
        Floating point tolerance as maximum deviation.

    Returns
    -------
    bool
    """
    nP = coords.shape[0]
    dA = distances_from_point(coords, coords[0])
    iB = np.argmax(dA)
    dB = distances_from_point(coords, coords[iB])
    iC = np.argmax(dA + dB)
    inds = np.array([0, iB, iC])
    d = np.zeros(3)
    d[:] = frame_of_plane(coords[inds, :])[1][:, 2]
    diffs = np.zeros(nP, dtype=coords.dtype)
    for i in prange(1, nP):
        diffs[i] = np.abs(coords[i] @ d)
    return diffs.max() <= tol


@njit(nogil=True, parallel=True, cache=__cache)
def move_points(coords: ndarray, p: ndarray) -> ndarray:
    res = np.zeros_like(coords)
    for i in prange(coords.shape[0]):
        res[i] = coords[i] + p
    return res



from typing import Union, Iterable, Generic, TypeVar, Optional
from copy import deepcopy

import numpy as np
from numpy import ndarray

from sigmaepsilon.core import classproperty
from sigmaepsilon.math import atleast3d, repeat
from sigmaepsilon.math.linalg import ReferenceFrame

from .akwrapper import AkWrapper
from ..typing import PolyDataProtocol, PointDataProtocol
from ..typing.abcakwrapper import ABC_AkWrapper
from .akwrapper import AwkwardLike

PointDataLike = TypeVar("PointDataLike", bound=PointDataProtocol)
PolyDataLike = TypeVar("PolyDataLike", bound=PolyDataProtocol)


__all__ = ["CellData"]


class CellData(Generic[PolyDataLike, PointDataLike], AkWrapper, ABC_AkWrapper):
    """
    A class to handle data related to the cells of a polygonal mesh.

    Technically this is a wrapper around an Awkward data object instance.

    If you are not a developer, you probably don't have to ever create any
    instance of this class, but since it operates in the background of every
    polygonal data structure, it is useful to understand how it works.

    Parameters
    ----------
    activity: numpy.ndarray, Optional
        1d boolean array describing the activity of the elements.
    t or thickness: numpy.ndarray, Optional
        1d float array of thicknesses. Only for 2d cells.
        Default is None.
    areas: numpy.ndarray, Optional
        1d float array of cross sectional areas. Only for 1d cells.
        Default is None.
    fields: dict, Optional
        Every value of this dictionary is added to the dataset.
        Default is `None`.
    frames: numpy.ndarray, Optional
        Coordinate axes representing cartesian coordinate frames.
        Default is None.
    topo: numpy.ndarray, Optional
        2d integer array representing node indices. Default is None.
    i: numpy.ndarray, Optional
        The (global) indices of the cells. Default is None.
    **kwargs: dict, Optional
        For every key and value pair where the value is a numpy array
        with a matching shape (has entries for all cells), the key
        is considered as a field and the value is added to the database.
    """

    _attr_map_ = {
        "nodes": "_nodes",  # node indices
        "frames": "_frames",  # coordinate frames
        "ndf": "_ndf",  # nodal distribution factors
        "id": "_id",  # global indices of the cells
        "areas": "_areas",  # areas of 1d cells
        "t": "_t",  # thicknesses for 2d cells
        "activity": "_activity",  # activity of the cells
    }

    def __init__(
        self,
        *args,
        wrap: Optional[Union[AwkwardLike, None]] = None,
        topo: Optional[Union[ndarray, None]] = None,
        fields: Optional[Union[dict, None]] = None,
        activity: Optional[Union[ndarray, None]] = None,
        frames: Optional[Union[ndarray, ReferenceFrame, None]] = None,
        areas: Optional[Union[ndarray, float, None]] = None,
        t: Optional[Union[ndarray, float, None]] = None,
        db: Optional[Union[AwkwardLike, None]] = None,
        i: Optional[Union[ndarray, None]] = None,
        **kwargs,
    ):
        fields = {} if fields is None else fields
        assert isinstance(fields, dict)
        if len(fields) > 0:
            attr_map = self._attr_map_
            fields = {attr_map.get(k, k): v for k, v in fields.items()}

        # cell indices
        if isinstance(i, ndarray):
            kwargs[self._dbkey_id_] = i

        if db is not None:
            wrap = db
        elif wrap is None:
            nodes = None
            if len(args) > 0:
                if isinstance(args[0], ndarray):
                    nodes = args[0]
            else:
                nodes = topo

            if isinstance(activity, ndarray):
                fields[self._dbkey_activity_] = activity

            if isinstance(nodes, ndarray):
                fields[self._dbkey_nodes_] = nodes
                N = nodes.shape[0]
                for k, v in kwargs.items():
                    if isinstance(v, ndarray):
                        if v.shape[0] == N:
                            fields[k] = v

            if isinstance(areas, np.ndarray):
                fields[self._dbkey_areas_] = areas

        super().__init__(*args, wrap=wrap, fields=fields, **kwargs)

        if self.db is not None:
            if frames is not None:
                if isinstance(frames, (ReferenceFrame, ndarray)):
                    self.frames = frames
                else:
                    msg = (
                        "'frames' must be a NumPy array, or a ",
                        "sigmaepsilon.math.linalg.ReferenceFrame instance.",
                    )
                    raise TypeError(msg)

            if t is not None:
                self.t = t

            if areas is not None:
                self.A = areas

    def __deepcopy__(self, memo: dict) -> "CellData":
        return self.__copy__(memo)

    def __copy__(self, memo: dict = None) -> "CellData":
        cls = type(self)
        is_deep = memo is not None

        if is_deep:
            copy_function = lambda x: deepcopy(x, memo)
        else:
            copy_function = lambda x: x

        db = copy_function(self.db)

        result = cls(db=db)
        if is_deep:
            memo[id(self)] = result

        result_dict = result.__dict__
        for k, v in self.__dict__.items():
            if not k in result_dict:
                setattr(result, k, copy_function(v))

        return result

    @classproperty
    def _dbkey_nodes_(cls) -> str:
        return cls._attr_map_["nodes"]

    @classproperty
    def _dbkey_frames_(cls) -> str:
        return cls._attr_map_["frames"]

    @classproperty
    def _dbkey_areas_(cls) -> str:
        return cls._attr_map_["areas"]

    @classproperty
    def _dbkey_thickness_(cls) -> str:
        return cls._attr_map_["t"]

    @classproperty
    def _dbkey_activity_(cls) -> str:
        return cls._attr_map_["activity"]

    @classproperty
    def _dbkey_ndf_(cls) -> str:
        return cls._attr_map_["ndf"]

    @classproperty
    def _dbkey_id_(cls) -> str:
        return cls._attr_map_["id"]

    @property
    def has_nodes(self) -> bool:
        return self._dbkey_nodes_ in self._wrapped.fields

    @property
    def has_id(self) -> bool:
        return self._dbkey_id_ in self._wrapped.fields

    @property
    def has_frames(self) -> bool:
        return self._dbkey_frames_ in self._wrapped.fields

    @property
    def has_thickness(self) -> bool:
        return self._dbkey_thickness_ in self._wrapped.fields

    @property
    def has_areas(self) -> bool:
        return self._dbkey_areas_ in self._wrapped.fields

    @property
    def db(self) -> AwkwardLike:
        return self._wrapped

    @property
    def fields(self) -> Iterable[str]:
        """Returns the fields in the database object."""
        return self._wrapped.fields

    @property
    def nodes(self) -> ndarray:
        """Returns the topology of the cells."""
        return self._wrapped[self._dbkey_nodes_].to_numpy()

    @nodes.setter
    def nodes(self, value: ndarray) -> None:
        """
        Sets the topology of the cells.

        Parameters
        ----------
        value: numpy.ndarray
            A 2d integer array.
        """
        assert isinstance(value, ndarray)
        self._wrapped[self._dbkey_nodes_] = value

    @property
    def frames(self) -> ndarray:
        """Returns local coordinate frames of the cells."""
        return self._wrapped[self._dbkey_frames_].to_numpy()

    @frames.setter
    def frames(self, value: Union[ReferenceFrame, ndarray]) -> None:
        """
        Sets local coordinate frames of the cells.

        Parameters
        ----------
        value: numpy.ndarray
            A 3d float array.
        """
        if isinstance(value, ReferenceFrame):
            frames = value.show()
        elif isinstance(value, ndarray):
            frames = value
        else:
            raise TypeError(f"Expected ndarray or FrameLike, got {type(value)}.")

        frames = atleast3d(frames)

        if len(frames) == 1:
            frames = repeat(frames[0], len(self._wrapped))
        else:
            assert len(frames) == len(self._wrapped)

        self._wrapped[self._dbkey_frames_] = frames

    @property
    def t(self) -> ndarray:
        """Returns the thicknesses of the cells."""
        return self._wrapped[self._dbkey_thickness_].to_numpy()

    @t.setter
    def t(self, value: Union[float, int, ndarray]):
        """Returns the thicknesses of the cells."""
        if isinstance(value, (int, float)):
            value = np.full(len(self), value)
        self._wrapped[self._dbkey_thickness_] = value

    @property
    def A(self) -> ndarray:
        """Returns the thicknesses of the cells."""
        return self._wrapped[self._dbkey_areas_].to_numpy()

    @A.setter
    def A(self, value: Union[float, int, ndarray]):
        """Returns the thicknesses of the cells."""
        if isinstance(value, (int, float)):
            value = np.full(len(self), value)
        self._wrapped[self._dbkey_areas_] = value

    @property
    def id(self) -> ndarray:
        """Returns global indices of the cells."""
        return self._wrapped[self._dbkey_id_].to_numpy()

    @id.setter
    def id(self, value: ndarray) -> None:
        """
        Sets global indices of the cells.

        Parameters
        ----------
        value: numpy.ndarray
            An 1d integer array.
        """
        if isinstance(value, int):
            if len(self) == 1:
                value = np.array(
                    [
                        value,
                    ],
                    dtype=int,
                )
            else:
                raise ValueError(f"Expected an array, got {type(value)}")

        if not isinstance(value, ndarray):
            raise TypeError(f"Expected ndarray, got {type(value)}")

        self._wrapped[self._dbkey_id_] = value

    @property
    def activity(self) -> ndarray:
        """Returns a 1d boolean array of cell activity."""
        return self._wrapped[self._dbkey_activity_].to_numpy()

    @activity.setter
    def activity(self, value: ndarray):
        """
        Sets cell activity with a 1d boolean array.

        Parameters
        ----------
        value: numpy.ndarray
            An 1d bool array.
        """
        if isinstance(value, bool):
            value = np.full(len(self), value, dtype=bool)
        self._wrapped[self._dbkey_activity_] = value

    def set_nodal_distribution_factors(self, factors: ndarray, key: str = None) -> None:
        """
        Sets nodal distribution factors.

        Parameters
        ----------
        factors: numpy.ndarray
            A 3d float array. The length of the array must equal the number
            pf cells in the block.
        key: str, Optional
            A key used to store the values in the database. This makes you able
            to use more nodal distribution strategies in one model.
            If not specified, a default key is used.
        """
        if key is None:
            key = self.__class__._attr_map_[self._dbkey_ndf_]
        if len(factors) != len(self._wrapped):
            self._wrapped[key] = factors[self._wrapped.id]
        else:
            self._wrapped[key] = factors
